import tempfile
import textwrap

import click
from click import testing as cli_testing
import pytest
import yaml

import sky
from sky import backends
from sky import cli
from sky import exceptions
from sky import global_user_state
from sky.utils import common
from sky.utils import controller_utils
from sky.utils import db_utils


def test_job_nonexist_strategy():
    """Test the nonexist recovery strategy.

    This function is testing for the core functions on server side.
    """
    task_yaml = textwrap.dedent("""\
        resources:
            cloud: aws
            use_spot: true
            job_recovery: nonexist""")
    with tempfile.NamedTemporaryFile(mode='w') as f:
        f.write(task_yaml)
        f.flush()
        with pytest.raises(ValueError,
                           match='is not a valid jobs recovery strategy among'):
            task = sky.Task.from_yaml(f.name)
            task.validate()


@pytest.fixture
def _mock_db_conn(tmp_path, monkeypatch):
    # Create a temporary database file
    db_path = tmp_path / 'state_testing.db'

    # Create a new SQLiteConn instance
    db_conn = db_utils.SQLiteConn(str(db_path), global_user_state.create_table)

    # Monkeypatch the global database connection
    monkeypatch.setattr(global_user_state, '_DB', db_conn)


def _generate_tmp_yaml(tmp_path, filename: str) -> str:
    yaml_path = tmp_path / filename
    private_key_path = tmp_path / 'id_rsa'
    private_key_path.write_text('test')
    text = yaml.dump({
        'auth': {
            'ssh_user': 'ubuntu',
            'ssh_private_key': str(private_key_path)
        },
        'provider': {
            'module': 'sky.backends.aws'
        }
    })
    yaml_path.write_text(text)
    return str(yaml_path)


@pytest.fixture
def _mock_cluster_state(_mock_db_conn, tmp_path):
    assert 'state.db' not in global_user_state._DB.db_path
    # Mock an empty /tmp/cluster1.yaml using tmp_path

    handle = backends.CloudVmRayResourceHandle(
        cluster_name='test-cluster1',
        cluster_name_on_cloud='test-cluster1',
        cluster_yaml=_generate_tmp_yaml(tmp_path, 'cluster1.yaml'),
        launched_nodes=2,
        launched_resources=sky.Resources(sky.AWS(),
                                         instance_type='p3.2xlarge',
                                         region='us-east-1'),
    )
    global_user_state.add_or_update_cluster(
        'test-cluster1',
        handle,
        requested_resources={handle.launched_resources},
        ready=True)
    handle = backends.CloudVmRayResourceHandle(
        cluster_name='test-cluster2',
        cluster_name_on_cloud='test-cluster2',
        cluster_yaml=_generate_tmp_yaml(tmp_path, 'cluster2.yaml'),
        launched_nodes=1,
        launched_resources=sky.Resources(sky.GCP(),
                                         instance_type='a2-highgpu-4g',
                                         accelerators={'A100': 4},
                                         region='us-west1'),
    )
    global_user_state.add_or_update_cluster(
        'test-cluster2',
        handle,
        requested_resources={handle.launched_resources},
        ready=True)
    handle = backends.CloudVmRayResourceHandle(
        cluster_name='test-cluster3',
        cluster_name_on_cloud='test-cluster3',
        cluster_yaml=_generate_tmp_yaml(tmp_path, 'cluster3.yaml'),
        launched_nodes=4,
        launched_resources=sky.Resources(sky.Azure(),
                                         instance_type='Standard_D4s_v3',
                                         region='eastus'),
    )
    global_user_state.add_or_update_cluster(
        'test-cluster3',
        handle,
        requested_resources={handle.launched_resources},
        ready=False)


@pytest.fixture
def _mock_jobs_controller(_mock_db_conn, tmp_path):
    handle = backends.CloudVmRayResourceHandle(
        cluster_name=common.JOB_CONTROLLER_NAME,
        cluster_name_on_cloud=common.JOB_CONTROLLER_NAME,
        cluster_yaml=_generate_tmp_yaml(tmp_path, 'jobs_controller.yaml'),
        launched_nodes=1,
        launched_resources=sky.Resources(sky.AWS(),
                                         instance_type='m4.2xlarge',
                                         region='us-west-1'),
    )
    global_user_state.add_or_update_cluster(
        common.JOB_CONTROLLER_NAME,
        handle,
        requested_resources={handle.launched_resources},
        ready=True)


@pytest.fixture
def _mock_serve_controller(_mock_db_conn, tmp_path):
    yaml_path = _generate_tmp_yaml(tmp_path, 'serve_controller.yaml')
    handle = backends.CloudVmRayResourceHandle(
        cluster_name=common.SKY_SERVE_CONTROLLER_NAME,
        cluster_name_on_cloud=common.SKY_SERVE_CONTROLLER_NAME,
        cluster_yaml=yaml_path,
        launched_nodes=1,
        launched_resources=sky.Resources(sky.AWS(),
                                         instance_type='m4.2xlarge',
                                         region='us-west-1'),
        stable_internal_external_ips=[('1.2.3.4', '4.3.2.1')],
        stable_ssh_ports=[22],
    )
    global_user_state.add_or_update_cluster(
        common.SKY_SERVE_CONTROLLER_NAME,
        handle,
        requested_resources={handle.launched_resources},
        ready=True)


class TestWithEmptyDBSetup:

    @pytest.fixture(autouse=True)
    def setup(self, _mock_db_conn, mock_client_requests):
        pass

    def test_cancel_jobs(self):
        cli_runner = cli_testing.CliRunner()
        result = cli_runner.invoke(cli.jobs_cancel, ['-a', '-y'])
        assert result.exit_code == 1

        assert isinstance(result.exception, exceptions.ClusterNotUpError)
        assert controller_utils.Controllers.JOBS_CONTROLLER.value.default_hint_if_non_existent in str(
            result.exception), (result.exception, result.output,
                                result.exc_info)

    def test_logs_jobs(self):
        cli_runner = cli_testing.CliRunner()
        result = cli_runner.invoke(cli.jobs_logs, ['1'])
        assert result.exit_code == 1
        assert controller_utils.Controllers.JOBS_CONTROLLER.value.default_hint_if_non_existent in str(
            result.exception), (result.exception, result.output,
                                result.exc_info)

    def test_queue_jobs(self):
        cli_runner = cli_testing.CliRunner()
        result = cli_runner.invoke(cli.jobs_queue)
        assert result.exit_code == 0
        assert controller_utils.Controllers.JOBS_CONTROLLER.value.default_hint_if_non_existent in str(
            result.output), (result.exception, result.output, result.exc_info)

    @pytest.mark.timeout(60)
    def test_down_serve(self):
        cli_runner = cli_testing.CliRunner()
        result = cli_runner.invoke(cli.serve_down, ['-a', '-y'])
        assert result.exit_code == 1

        assert isinstance(result.exception, exceptions.ClusterNotUpError)
        assert (controller_utils.Controllers.SKY_SERVE_CONTROLLER.value.
                default_hint_if_non_existent
                in str(result.exception)), (result.exception, result.output,
                                            result.exc_info)

    def test_logs_serve(self):
        cli_runner = cli_testing.CliRunner()
        result = cli_runner.invoke(cli.serve_logs, ['test', '--controller'])
        assert controller_utils.Controllers.SKY_SERVE_CONTROLLER.value.default_hint_if_non_existent in str(
            result.exception), (result.exception, result.output,
                                result.exc_info)

    def test_status_serve(self):
        cli_runner = cli_testing.CliRunner()
        result = cli_runner.invoke(cli.serve_status)
        assert result.exit_code == 0
        assert controller_utils.Controllers.SKY_SERVE_CONTROLLER.value.default_hint_if_non_existent in str(
            result.output), (result.exception, result.output, result.exc_info)


class TestJobsOperations:
    """Test operations for managed jobs."""

    @pytest.fixture(autouse=True)
    def setup(self, _mock_db_conn, _mock_cluster_state, _mock_jobs_controller,
              mock_controller_accessible, mock_job_table_one_job,
              mock_client_requests):
        pass

    def test_stop_jobs_controller(self):
        cli_runner = cli_testing.CliRunner()
        result = cli_runner.invoke(cli.stop, [common.JOB_CONTROLLER_NAME])
        assert result.exit_code == click.UsageError.exit_code
        assert (f'Stopping controller(s) \'{common.JOB_CONTROLLER_NAME}\' is '
                'currently not supported' in result.output)

        result = cli_runner.invoke(cli.stop, ['sky-jobs-con*'])
        assert not result.exception
        assert 'Cluster(s) not found' in result.output

        result = cli_runner.invoke(cli.stop, ['-a'], input='n\n')
        assert isinstance(result.exception, SystemExit)
        assert 'Aborted' in result.output

    def test_autostop_jobs_controller(self):
        cli_runner = cli_testing.CliRunner()
        result = cli_runner.invoke(cli.autostop, [common.JOB_CONTROLLER_NAME])
        assert result.exit_code == click.UsageError.exit_code
        assert ('Scheduling autostop on controller(s) '
                f'\'{common.JOB_CONTROLLER_NAME}\' is currently not supported'
                in result.output)

        result = cli_runner.invoke(cli.autostop, ['sky-jobs-con*'])
        assert not result.exception
        assert 'Cluster(s) not found' in result.output

        result = cli_runner.invoke(cli.autostop, ['-a'], input='n\n')
        assert isinstance(result.exception, SystemExit)
        assert 'Aborted' in result.output

    def test_cancel_on_jobs_controller(self):
        cli_runner = cli_testing.CliRunner()
        result = cli_runner.invoke(cli.cancel,
                                   [common.JOB_CONTROLLER_NAME, '-a'])
        assert result.exit_code == click.UsageError.exit_code
        assert 'Cancelling the jobs controller\'s jobs is not allowed.' in str(
            result.output)

    def test_down_jobs_controller_no_job(self, mock_job_table_no_job,
                                         mock_client_requests):
        cli_runner = cli_testing.CliRunner()
        result = cli_runner.invoke(cli.down, [common.JOB_CONTROLLER_NAME],
                                   input='n')
        assert 'WARNING: Tearing down the managed jobs controller.' in result.output, (
            result.exception, result.output, result.exc_info)
        assert isinstance(result.exception,
                          SystemExit), (result.exception, result.output)

    def test_down_jobs_controller_one_job(self):
        cli_runner = cli_testing.CliRunner()
        result = cli_runner.invoke(cli.down, [common.JOB_CONTROLLER_NAME],
                                   input='n')
        assert 'WARNING: Tearing down the managed jobs controller.' in result.output, (
            result.exception, result.output, result.exc_info)
        assert isinstance(result.exception, exceptions.NotSupportedError)

        result = cli_runner.invoke(cli.down, ['sky-jobs-con*'])
        assert not result.exception
        assert 'Cluster(s) not found' in result.output

        result = cli_runner.invoke(cli.down, ['sky-jobs-con*', '-p'])
        assert not result.exception
        assert 'Cluster(s) not found' in result.output

        result = cli_runner.invoke(cli.down, ['-a'], input='n\n')
        assert isinstance(result.exception, SystemExit), result.exception
        assert 'Aborted' in result.output

        result = cli_runner.invoke(cli.down, ['-ap'], input='n\n')
        assert isinstance(result.exception, SystemExit)
        assert 'Aborted' in result.output


class TestServeOperations:
    """Test operations for services."""

    @pytest.fixture(autouse=True)
    def setup(self, _mock_db_conn, _mock_cluster_state, _mock_serve_controller,
              mock_client_requests):
        pass

    def test_stop_serve_controller(self,):
        cli_runner = cli_testing.CliRunner()
        result = cli_runner.invoke(cli.stop, [common.SKY_SERVE_CONTROLLER_NAME])
        assert result.exit_code == click.UsageError.exit_code
        assert (
            f'Stopping controller(s) \'{common.SKY_SERVE_CONTROLLER_NAME}\' is '
            'currently not supported' in result.output)

        result = cli_runner.invoke(cli.stop, ['sky-serve-con*'])
        assert not result.exception
        assert 'Cluster(s) not found' in result.output

        result = cli_runner.invoke(cli.stop, ['-a'], input='n\n')
        assert isinstance(result.exception, SystemExit)
        assert 'Aborted' in result.output

    def test_autostop_serve_controller(self):
        cli_runner = cli_testing.CliRunner()
        result = cli_runner.invoke(cli.autostop,
                                   [common.SKY_SERVE_CONTROLLER_NAME])
        assert result.exit_code == click.UsageError.exit_code
        assert (
            'Scheduling autostop on controller(s) '
            f'\'{common.SKY_SERVE_CONTROLLER_NAME}\' is currently not supported'
            in result.output)

        result = cli_runner.invoke(cli.autostop, ['sky-serve-con*'])
        assert not result.exception
        assert 'Cluster(s) not found' in result.output

        result = cli_runner.invoke(cli.autostop, ['-a'], input='n\n')
        assert isinstance(result.exception, SystemExit)
        assert 'Aborted' in result.output

    def test_cancel_on_serve_controller(self):
        cli_runner = cli_testing.CliRunner()
        result = cli_runner.invoke(cli.cancel,
                                   [common.SKY_SERVE_CONTROLLER_NAME, '-a'])
        assert result.exit_code == click.UsageError.exit_code
        assert 'Cancelling the sky serve controller\'s jobs is not allowed.' in str(
            result.output)

    def test_down_serve_controller_one_service(self, mock_controller_accessible,
                                               mock_services_one_service):
        cli_runner = cli_testing.CliRunner()

        result = cli_runner.invoke(cli.down, [common.SKY_SERVE_CONTROLLER_NAME],
                                   input='n')
        assert isinstance(result.exception, exceptions.NotSupportedError)

        result = cli_runner.invoke(cli.down, ['sky-serve-con*'])
        assert not result.exception
        assert 'Cluster(s) not found' in result.output

        result = cli_runner.invoke(cli.down, ['sky-serve-con*', '-p'])
        assert not result.exception
        assert 'Cluster(s) not found' in result.output

        result = cli_runner.invoke(cli.down, ['-a'], input='n\n')
        assert isinstance(result.exception, SystemExit), result.exception
        assert 'Aborted' in result.output

        result = cli_runner.invoke(cli.down, ['-ap'], input='n\n')
        assert isinstance(result.exception, SystemExit)
        assert 'Aborted' in result.output

    def test_down_serve_controller_no_service(self, mock_controller_accessible,
                                              mock_services_no_service):
        cli_runner = cli_testing.CliRunner()
        result = cli_runner.invoke(cli.down, [common.SKY_SERVE_CONTROLLER_NAME],
                                   input='n')
        assert 'Terminate sky serve controller:' in result.output, (
            result.exception, result.output, result.exc_info)
        assert isinstance(result.exception,
                          SystemExit), (result.exception, result.output)



"""Kubernetes adaptors"""
import logging
import os
from typing import Any, Callable, Optional, Set

from sky.adaptors import common
from sky.sky_logging import set_logging_level
from sky.utils import annotations
from sky.utils import common_utils
from sky.utils import ux_utils

_IMPORT_ERROR_MESSAGE = ('Failed to import dependencies for Kubernetes. '
                         'Try running: pip install "skypilot[kubernetes]"')
kubernetes = common.LazyImport('kubernetes',
                               import_error_message=_IMPORT_ERROR_MESSAGE)
urllib3 = common.LazyImport('urllib3',
                            import_error_message=_IMPORT_ERROR_MESSAGE)

# Timeout to use for API calls
API_TIMEOUT = 5

DEFAULT_IN_CLUSTER_REGION = 'in-cluster'
# The name for the environment variable that stores the in-cluster context name
# for Kubernetes clusters. This is used to associate a name with the current
# context when running with in-cluster auth. If not set, the context name is
# set to DEFAULT_IN_CLUSTER_REGION.
IN_CLUSTER_CONTEXT_NAME_ENV_VAR = 'SKYPILOT_IN_CLUSTER_CONTEXT_NAME'


def _decorate_methods(obj: Any, decorator: Callable, decoration_type: str):
    for attr_name in dir(obj):
        attr = getattr(obj, attr_name)
        # Skip methods starting with '__' since they are invoked through one
        # of the main methods, which are already decorated.
        if callable(attr) and not attr_name.startswith('__'):
            decorated_types: Set[str] = getattr(attr, '_sky_decorator_types',
                                                set())
            if decoration_type not in decorated_types:
                decorated_attr = decorator(attr)
                decorated_attr._sky_decorator_types = (  # pylint: disable=protected-access
                    decorated_types | {decoration_type})
                setattr(obj, attr_name, decorated_attr)
    return obj


def _api_logging_decorator(logger: str, level: int):
    """Decorator to set logging level for API calls.

    This is used to suppress the verbose logging from urllib3 when calls to the
    Kubernetes API timeout.
    """

    def decorated_api(api):

        def wrapped(*args, **kwargs):
            obj = api(*args, **kwargs)
            _decorate_methods(obj, set_logging_level(logger, level), 'api_log')
            return obj

        return wrapped

    return decorated_api


def _load_config(context: Optional[str] = None):
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

    def _load_config_from_kubeconfig(context: Optional[str] = None):
        try:
            kubernetes.config.load_kube_config(context=context)
        except kubernetes.config.config_exception.ConfigException as e:
            suffix = common_utils.format_exception(e, use_bracket=True)
            context_name = '(current-context)' if context is None else context
            # Check if exception was due to no current-context
            if 'Expected key current-context' in str(e):
                err_str = ('Failed to load Kubernetes configuration for '
                           f'{context_name!r}. '
                           'Kubeconfig does not contain any valid context(s).'
                           f'\n{suffix}\n'
                           '    If you were running a local Kubernetes '
                           'cluster, run `sky local up` to start the cluster.')
            else:
                kubeconfig_path = os.environ.get('KUBECONFIG', '~/.kube/config')
                err_str = (
                    f'Failed to load Kubernetes configuration for '
                    f'{context_name!r}. Please check if your kubeconfig file '
                    f'exists at {kubeconfig_path} and is valid.\n{suffix}')
            err_str += '\nTo disable Kubernetes for SkyPilot: run `sky check`.'
            if context is None:  # kubernetes defaults to current-context.
                err_str += (
                    '\nHint: Kubernetes attempted to query the current-context '
                    'set in kubeconfig. Check if the current-context is valid.')
            with ux_utils.print_exception_no_traceback():
                raise ValueError(err_str) from None

    if context == in_cluster_context_name() or context is None:
        try:
            # Load in-cluster config if running in a pod and context is None.
            # Kubernetes set environment variables for service discovery do not
            # show up in SkyPilot tasks. For now, we work around by using
            # DNS name instead of environment variables.
            # See issue: https://github.com/skypilot-org/skypilot/issues/2287
            os.environ['KUBERNETES_SERVICE_HOST'] = 'kubernetes.default.svc'
            os.environ['KUBERNETES_SERVICE_PORT'] = '443'
            kubernetes.config.load_incluster_config()
        except kubernetes.config.config_exception.ConfigException:
            _load_config_from_kubeconfig()
    else:
        _load_config_from_kubeconfig(context)


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
def core_api(context: Optional[str] = None):
    _load_config(context)
    return kubernetes.client.CoreV1Api()


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
def auth_api(context: Optional[str] = None):
    _load_config(context)
    return kubernetes.client.RbacAuthorizationV1Api()


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
def networking_api(context: Optional[str] = None):
    _load_config(context)
    return kubernetes.client.NetworkingV1Api()


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
def custom_objects_api(context: Optional[str] = None):
    _load_config(context)
    return kubernetes.client.CustomObjectsApi()


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='global')
def node_api(context: Optional[str] = None):
    _load_config(context)
    return kubernetes.client.NodeV1Api()


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
def apps_api(context: Optional[str] = None):
    _load_config(context)
    return kubernetes.client.AppsV1Api()


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
def batch_api(context: Optional[str] = None):
    _load_config(context)
    return kubernetes.client.BatchV1Api()


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
def api_client(context: Optional[str] = None):
    _load_config(context)
    return kubernetes.client.ApiClient()


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
def watch(context: Optional[str] = None):
    _load_config(context)
    return kubernetes.watch.Watch()


def api_exception():
    return kubernetes.client.rest.ApiException


def config_exception():
    return kubernetes.config.config_exception.ConfigException


def max_retry_error():
    return urllib3.exceptions.MaxRetryError


def stream():
    return kubernetes.stream.stream


def in_cluster_context_name() -> Optional[str]:
    """Returns the name of the in-cluster context from the environment.

    If the environment variable is not set, returns the default in-cluster
    context name.
    """
    return (os.environ.get(IN_CLUSTER_CONTEXT_NAME_ENV_VAR) or
            DEFAULT_IN_CLUSTER_REGION)



"""OCI instance provisioning.

History:
 - Hysun He (hysun.he@oracle.com) @ Oct.16, 2024: Initial implementation
 - Hysun He (hysun.he@oracle.com) @ Nov.13, 2024: Implement open_ports
   and cleanup_ports for supporting SkyServe.
"""

import copy
from datetime import datetime
import time
import typing
from typing import Any, Dict, List, Optional

from sky import exceptions
from sky import sky_logging
from sky.adaptors import oci as oci_adaptor
from sky.clouds.utils import oci_utils
from sky.provision import common
from sky.provision import constants
from sky.provision.oci import query_utils
from sky.provision.oci.query_utils import query_helper
from sky.utils import common_utils
from sky.utils import ux_utils

if typing.TYPE_CHECKING:
    from sky.utils import status_lib

logger = sky_logging.init_logger(__name__)


@query_utils.debug_enabled(logger)
@common_utils.retry
def query_instances(
    cluster_name_on_cloud: str,
    provider_config: Optional[Dict[str, Any]] = None,
    non_terminated_only: bool = True,
) -> Dict[str, Optional['status_lib.ClusterStatus']]:
    """Query instances.

    Returns a dictionary of instance IDs and status.

    A None status means the instance is marked as "terminated"
    or "terminating".
    """
    assert provider_config is not None, cluster_name_on_cloud
    region = provider_config['region']

    status_map = oci_utils.oci_config.STATE_MAPPING_OCI_TO_SKY
    statuses: Dict[str, Optional['status_lib.ClusterStatus']] = {}
    filters = {constants.TAG_RAY_CLUSTER_NAME: cluster_name_on_cloud}

    instances = _get_filtered_nodes(region, filters)
    for node in instances:
        vm_status = node['status']
        sky_status = status_map[vm_status]
        if non_terminated_only and sky_status is None:
            continue
        statuses[node['inst_id']] = sky_status

    return statuses


@query_utils.debug_enabled(logger)
def run_instances(region: str, cluster_name_on_cloud: str,
                  config: common.ProvisionConfig) -> common.ProvisionRecord:
    """Start instances with bootstrapped configuration."""
    tags = dict(sorted(copy.deepcopy(config.tags).items()))

    start_time = round(time.time() * 1000)
    filters = {constants.TAG_RAY_CLUSTER_NAME: cluster_name_on_cloud}

    # Starting stopped nodes if resume_stopped_nodes=True
    resume_instances = []
    if config.resume_stopped_nodes:
        logger.debug('Checking existing stopped nodes.')

        existing_instances = _get_filtered_nodes(region, filters)
        if len(existing_instances) > config.count:
            raise RuntimeError(
                'The number of pending/running/stopped/stopping '
                f'instances combined ({len(existing_instances)}) in '
                f'cluster "{cluster_name_on_cloud}" is greater than the '
                f'number requested by the user ({config.count}). '
                'This is likely a resource leak. '
                'Use "sky down" to terminate the cluster.')

        # pylint: disable=line-too-long
        logger.debug(
            f'run_instances: Found {[inst["name"] for inst in existing_instances]} '
            'existing instances in cluster.')
        existing_instances.sort(key=lambda x: x['name'])

        stopped_instances = []
        for existing_node in existing_instances:
            if existing_node['status'] == 'STOPPING':
                query_helper.wait_instance_until_status(
                    region, existing_node['inst_id'], 'STOPPED')
                stopped_instances.append(existing_node)
            elif existing_node['status'] == 'STOPPED':
                stopped_instances.append(existing_node)
            elif existing_node['status'] in ('PROVISIONING', 'STARTING',
                                             'RUNNING'):
                resume_instances.append(existing_node)

        for stopped_node in stopped_instances:
            stopped_node_id = stopped_node['inst_id']
            instance_action_response = query_helper.start_instance(
                region, stopped_node_id)

            starting_inst = instance_action_response.data
            resume_instances.append({
                'inst_id': starting_inst.id,
                'name': starting_inst.display_name,
                'ad': starting_inst.availability_domain,
                'compartment': starting_inst.compartment_id,
                'status': starting_inst.lifecycle_state,
                'oci_tags': starting_inst.freeform_tags,
            })
    # end if config.resume_stopped_nodes

    # Try get head id from the existing instances
    head_instance_id = _get_head_instance_id(resume_instances)
    logger.debug(f'Check existing head node: {head_instance_id}')

    # Let's create additional new nodes (if neccessary)
    to_start_count = config.count - len(resume_instances)
    created_instances = []
    node_config = config.node_config
    if to_start_count > 0:
        compartment = query_helper.find_compartment(region)
        vcn = query_helper.find_create_vcn_subnet(region)

        ocpu_count = 0
        vcpu_str = node_config['VCPUs']
        instance_type_str = node_config['InstanceType']

        if vcpu_str is not None and vcpu_str != 'None':
            if instance_type_str.startswith(
                    f'{oci_utils.oci_config.VM_PREFIX}.A'):
                # For ARM cpu, 1*ocpu = 1*vcpu
                ocpu_count = round(float(vcpu_str))
            else:
                # For Intel / AMD cpu, 1*ocpu = 2*vcpu
                ocpu_count = round(float(vcpu_str) / 2)
        ocpu_count = 1 if (ocpu_count > 0 and ocpu_count < 1) else ocpu_count

        machine_shape_config = None
        if ocpu_count > 0:
            mem = node_config['MemoryInGbs']
            if mem is not None and mem != 'None':
                # pylint: disable=line-too-long
                machine_shape_config = oci_adaptor.oci.core.models.LaunchInstanceShapeConfigDetails(
                    ocpus=ocpu_count, memory_in_gbs=mem)
            else:
                # pylint: disable=line-too-long
                machine_shape_config = oci_adaptor.oci.core.models.LaunchInstanceShapeConfigDetails(
                    ocpus=ocpu_count)

        preempitible_config = (
            oci_adaptor.oci.core.models.PreemptibleInstanceConfigDetails(
                preemption_action=oci_adaptor.oci.core.models.
                TerminatePreemptionAction(type='TERMINATE',
                                          preserve_boot_volume=False))
            if node_config['Preemptible'] else None)

        batch_id = datetime.now().strftime('%Y%m%d%H%M%S')

        vm_tags_head = {
            **tags,
            **constants.HEAD_NODE_TAGS,
            constants.TAG_RAY_CLUSTER_NAME: cluster_name_on_cloud,
            'sky_spot_flag': str(node_config['Preemptible']).lower(),
        }
        vm_tags_worker = {
            **tags,
            **constants.WORKER_NODE_TAGS,
            constants.TAG_RAY_CLUSTER_NAME: cluster_name_on_cloud,
            'sky_spot_flag': str(node_config['Preemptible']).lower(),
        }

        for seq in range(1, to_start_count + 1):
            if head_instance_id is None:
                vm_tags = vm_tags_head
                node_type = constants.HEAD_NODE_TAGS[
                    constants.TAG_RAY_NODE_KIND]
            else:
                vm_tags = vm_tags_worker
                node_type = constants.WORKER_NODE_TAGS[
                    constants.TAG_RAY_NODE_KIND]

            launch_instance_response = query_helper.launch_instance(
                region,
                oci_adaptor.oci.core.models.LaunchInstanceDetails(
                    availability_domain=node_config['AvailabilityDomain'],
                    compartment_id=compartment,
                    shape=instance_type_str,
                    display_name=
                    f'{cluster_name_on_cloud}_{node_type}_{batch_id}_{seq}',
                    freeform_tags=vm_tags,
                    metadata={
                        'ssh_authorized_keys': node_config['AuthorizedKey']
                    },
                    source_details=oci_adaptor.oci.core.models.
                    InstanceSourceViaImageDetails(
                        source_type='image',
                        image_id=node_config['ImageId'],
                        boot_volume_size_in_gbs=node_config['BootVolumeSize'],
                        boot_volume_vpus_per_gb=int(
                            node_config['BootVolumePerf']),
                    ),
                    create_vnic_details=oci_adaptor.oci.core.models.
                    CreateVnicDetails(
                        assign_public_ip=True,
                        subnet_id=vcn,
                    ),
                    shape_config=machine_shape_config,
                    preemptible_instance_config=preempitible_config,
                ))

            new_inst = launch_instance_response.data
            if head_instance_id is None:
                head_instance_id = new_inst.id
                logger.debug(f'New head node: {head_instance_id}')

            created_instances.append({
                'inst_id': new_inst.id,
                'name': new_inst.display_name,
                'ad': new_inst.availability_domain,
                'compartment': new_inst.compartment_id,
                'status': new_inst.lifecycle_state,
                'oci_tags': new_inst.freeform_tags,
            })
        # end for loop
    # end if to_start_count > 0:...

    for inst in (resume_instances + created_instances):
        logger.debug(f'Provisioning for node {inst["name"]}')
        query_helper.wait_instance_until_status(region, inst['inst_id'],
                                                'RUNNING')
        logger.debug(f'Instance {inst["name"]} is RUNNING.')

    total_time = round(time.time() * 1000) - start_time
    logger.debug('Total time elapsed: {0} milli-seconds.'.format(total_time))

    assert head_instance_id is not None, head_instance_id

    # Format: TenancyPrefix:AvailabilityDomain, e.g. bxtG:US-SANJOSE-1-AD-1
    _, ad = str(node_config['AvailabilityDomain']).split(':', maxsplit=1)
    return common.ProvisionRecord(
        provider_name='oci',
        region=region,
        zone=ad,
        cluster_name=cluster_name_on_cloud,
        head_instance_id=head_instance_id,
        created_instance_ids=[n['inst_id'] for n in created_instances],
        resumed_instance_ids=[n['inst_id'] for n in resume_instances],
    )


@query_utils.debug_enabled(logger)
def stop_instances(
    cluster_name_on_cloud: str,
    provider_config: Dict[str, Any],
    worker_only: bool = False,
) -> None:
    """Stop running instances."""
    # pylint: disable=line-too-long
    assert provider_config is not None, (cluster_name_on_cloud, provider_config)

    region = provider_config['region']
    tag_filters = {constants.TAG_RAY_CLUSTER_NAME: cluster_name_on_cloud}
    if worker_only:
        tag_filters[constants.TAG_RAY_NODE_KIND] = 'worker'

    nodes = _get_filtered_nodes(region, tag_filters)
    for node in nodes:
        query_helper.stop_instance(region, node['inst_id'])


@query_utils.debug_enabled(logger)
def terminate_instances(
    cluster_name_on_cloud: str,
    provider_config: Dict[str, Any],
    worker_only: bool = False,
) -> None:
    """Terminate running or stopped instances."""
    region = provider_config['region']
    tag_filters = {constants.TAG_RAY_CLUSTER_NAME: cluster_name_on_cloud}
    if worker_only:
        tag_filters[constants.TAG_RAY_NODE_KIND] = 'worker'
    query_helper.terminate_instances_by_tags(tag_filters, region)


@query_utils.debug_enabled(logger)
def open_ports(
    cluster_name_on_cloud: str,
    ports: List[str],
    provider_config: Optional[Dict[str, Any]] = None,
) -> None:
    """Open ports for inbound traffic."""
    assert provider_config is not None, cluster_name_on_cloud
    region = provider_config['region']
    query_helper.create_nsg_rules(region=region,
                                  cluster_name=cluster_name_on_cloud,
                                  ports=ports)


@query_utils.debug_enabled(logger)
def cleanup_ports(
    cluster_name_on_cloud: str,
    ports: List[str],
    provider_config: Optional[Dict[str, Any]] = None,
) -> None:
    """Delete any opened ports."""
    assert provider_config is not None, cluster_name_on_cloud
    region = provider_config['region']
    del ports
    query_helper.remove_cluster_nsg(region=region,
                                    cluster_name=cluster_name_on_cloud)


@query_utils.debug_enabled(logger)
def wait_instances(region: str, cluster_name_on_cloud: str,
                   state: Optional['status_lib.ClusterStatus']) -> None:
    del region, cluster_name_on_cloud, state
    # We already wait for the instances to be running in run_instances.
    # We can not implement the wait logic here because the provisioning
    # instances are not retrieveable by the QL 'query instance resources ...'.


@query_utils.debug_enabled(logger)
def get_cluster_info(
    region: str,
    cluster_name_on_cloud: str,
    provider_config: Optional[Dict[str, Any]] = None,
) -> common.ClusterInfo:
    """Get the metadata of instances in a cluster."""
    filters = {constants.TAG_RAY_CLUSTER_NAME: cluster_name_on_cloud}
    running_instances = _get_filtered_nodes(region, filters)

    instances = {}
    for running_instance in running_instances:
        inst = _get_inst_obj_with_ip(region, running_instance)
        instances[inst['id']] = [
            common.InstanceInfo(
                instance_id=inst['id'],
                internal_ip=inst['internal_ip'],
                external_ip=inst['external_ip'],
                tags=inst['tags'],
            )
        ]

    instances = dict(sorted(instances.items(), key=lambda x: x[0]))
    logger.debug(f'Cluster info: {instances}')

    head_instance_id = _get_head_instance_id(running_instances)
    logger.debug(f'Head instance id is {head_instance_id}')

    return common.ClusterInfo(
        provider_name='oci',
        head_instance_id=head_instance_id,
        instances=instances,
        provider_config=provider_config,
    )


def _get_filtered_nodes(region: str,
                        tag_filters: Dict[str, str]) -> List[Dict[str, Any]]:
    return_nodes = []

    try:
        insts = query_helper.query_instances_by_tags(tag_filters, region)
    except oci_adaptor.oci.exceptions.ServiceError as e:
        with ux_utils.print_exception_no_traceback():
            raise exceptions.ClusterStatusFetchingError(
                f'Failed to query status for OCI cluster {tag_filters}.'
                'Details: '
                f'{common_utils.format_exception(e, use_bracket=True)}')

    for inst in insts:
        inst_id = inst.identifier
        return_nodes.append({
            'inst_id': inst_id,
            'name': inst.display_name,
            'ad': inst.availability_domain,
            'compartment': inst.compartment_id,
            'status': inst.lifecycle_state,
            'oci_tags': inst.freeform_tags,
        })

    return return_nodes


def _get_inst_obj_with_ip(region: str, inst_info: Dict[str,
                                                       Any]) -> Dict[str, Any]:
    get_vnic_response = query_helper.get_instance_primary_vnic(
        region, inst_info)
    internal_ip = get_vnic_response.private_ip
    external_ip = get_vnic_response.public_ip
    if external_ip is None:
        external_ip = internal_ip

    return {
        'id': inst_info['inst_id'],
        'name': inst_info['name'],
        'external_ip': external_ip,
        'internal_ip': internal_ip,
        'tags': inst_info['oci_tags'],
        'status': inst_info['status'],
    }


def _get_head_instance_id(instances: List[Dict[str, Any]]) -> Optional[str]:
    head_instance_id = None
    head_node_tags = tuple(constants.HEAD_NODE_TAGS.items())
    for inst in instances:
        is_matched = True
        for k, v in head_node_tags:
            if (k, v) not in inst['oci_tags'].items():
                is_matched = False
                break
        if is_matched:
            if head_instance_id is not None:
                logger.warning(
                    'There are multiple head nodes in the cluster '
                    f'(current head instance id: {head_instance_id}, '
                    f'newly discovered id: {inst["inst_id"]}. It is likely '
                    f'that something goes wrong.')
                # Don't break here so that we can continue to check and
                # warn user about duplicate head instance issue so that
                # user can take further action on the abnormal cluster.

            head_instance_id = inst['inst_id']

    return head_instance_id



"""Service manager factory
"""

from sky.provision.vsphere.common import service_manager as service_manager_lib


class ServiceManagerFactory(object):
    """Factory class for getting service manager for a management node.
    """

    service_manager = None

    @classmethod
    def get_service_manager(cls, server, username, password, skip_verification):
        service_manager = service_manager_lib.ServiceManager(
            server, username, password, skip_verification)
        service_manager.connect()
        return service_manager



"""Interfaces: clouds, regions, and zones.

clouds.Cloud is lightweight stateless objects. SkyPilot may create multiple such
objects; therefore, subclasses should take care to make methods inexpensive to
call, and should not store heavy state. If state needs to be queried from the
cloud provider and cached, create a module in sky/clouds/utils/ and probably add
caches for the return value (e.g., sky/clouds/utils/gcp_utils), so they can be
reused across cloud object creation.
"""
import collections
import enum
import math
import typing
from typing import Dict, Iterable, Iterator, List, Optional, Set, Tuple, Union

from typing_extensions import assert_never

from sky import exceptions
from sky import skypilot_config
from sky.clouds import service_catalog
from sky.utils import log_utils
from sky.utils import resources_utils
from sky.utils import timeline
from sky.utils import ux_utils

if typing.TYPE_CHECKING:
    from sky import resources as resources_lib
    from sky.utils import status_lib


class CloudImplementationFeatures(enum.Enum):
    """Features that might not be implemented for all clouds.

    Used by Cloud.check_features_are_supported().

    NOTE: If any new feature is added, please check and update
    _cloud_unsupported_features in all clouds to make sure the
    check_features_are_supported() works as expected.
    """
    STOP = 'stop'
    MULTI_NODE = 'multi-node'
    CLONE_DISK_FROM_CLUSTER = 'clone_disk_from_cluster'
    IMAGE_ID = 'image_id'
    DOCKER_IMAGE = 'docker_image'
    SPOT_INSTANCE = 'spot_instance'
    CUSTOM_DISK_TIER = 'custom_disk_tier'
    OPEN_PORTS = 'open_ports'
    STORAGE_MOUNTING = 'storage_mounting'
    HOST_CONTROLLERS = 'host_controllers'  # Can run jobs/serve controllers
    HIGH_AVAILABILITY_CONTROLLERS = ('high_availability_controllers'
                                    )  # Controller can auto-restart
    AUTO_TERMINATE = 'auto_terminate'  # Pod/VM can stop or down itself
    AUTOSTOP = 'autostop'  # Pod/VM can stop itself
    AUTODOWN = 'autodown'  # Pod/VM can down itself


# Use str, enum.Enum to allow CloudCapability to be used as a string.
class CloudCapability(str, enum.Enum):
    # Compute capability.
    COMPUTE = 'compute'
    # Storage capability.
    STORAGE = 'storage'


ALL_CAPABILITIES = [CloudCapability.COMPUTE, CloudCapability.STORAGE]


class Region(collections.namedtuple('Region', ['name'])):
    """A region."""
    name: str
    zones: Optional[List['Zone']] = None

    def set_zones(self, zones: List['Zone']):
        self.zones = zones
        for zone in self.zones:
            zone.region = self
        return self


class Zone(collections.namedtuple('Zone', ['name'])):
    """A zone, typically grouped under a region."""
    name: str
    region: Region


class ProvisionerVersion(enum.Enum):
    """The version of the provisioner.

    1: [Deprecated] ray node provider based implementation
    2: [Deprecated] ray node provider for provisioning and SkyPilot provisioner
    for stopping and termination
    3: SkyPilot provisioner for both provisioning and stopping
    """
    RAY_AUTOSCALER = 1
    RAY_PROVISIONER_SKYPILOT_TERMINATOR = 2
    SKYPILOT = 3

    def __ge__(self, other):
        return self.value >= other.value


class StatusVersion(enum.Enum):
    """The version of the status query.

    1: [Deprecated] cloud-CLI based implementation
    2: SkyPilot provisioner based implementation
    """
    CLOUD_CLI = 1
    SKYPILOT = 2

    def __ge__(self, other):
        return self.value >= other.value


class OpenPortsVersion(enum.Enum):
    """The version of the open ports implementation.

    1: Open ports on launching of the cluster only, cannot be modified after
    provisioning of the cluster. This is for clouds like RunPod which only
    accepts port argument on VM creation API, and requires Web GUI and an VM
    restart to update ports. We currently do not support this.
    2: Open ports after provisioning of the cluster, updatable. This is for most
    of the cloud providers which allow opening ports using an programmable API
    and won't affect the running VMs.
    """
    LAUNCH_ONLY = 'LAUNCH ONLY'
    UPDATABLE = 'UPDATABLE'

    def __le__(self, other):
        versions = list(OpenPortsVersion)
        return versions.index(self) <= versions.index(other)


class Cloud:
    """A cloud provider."""

    _REPR = '<Cloud>'
    _DEFAULT_DISK_TIER = resources_utils.DiskTier.MEDIUM
    _BEST_DISK_TIER = resources_utils.DiskTier.ULTRA
    _SUPPORTED_DISK_TIERS = {resources_utils.DiskTier.BEST}
    _SUPPORTS_SERVICE_ACCOUNT_ON_REMOTE = False

    # The version of provisioner and status query. This is used to determine
    # the code path to use for each cloud in the backend.
    # NOTE: new clouds being added should use the latest version, i.e. SKYPILOT.
    PROVISIONER_VERSION = ProvisionerVersion.RAY_AUTOSCALER
    STATUS_VERSION = StatusVersion.CLOUD_CLI
    OPEN_PORTS_VERSION = OpenPortsVersion.UPDATABLE

    @classmethod
    def max_cluster_name_length(cls) -> Optional[int]:
        """Returns the maximum length limit of a cluster name.

        This method is used by check_cluster_name_is_valid() to check if the
        cluster name is too long.

        None means no limit.
        """
        return None

    @classmethod
    def supports_service_account_on_remote(cls) -> bool:
        """Returns whether the cloud supports service account on remote cluster.

        This method is used by backend_utils.write_cluster_config() to decide
        whether to upload user's local cloud credential files to the remote
        cluster.

        If a cloud supports service account on remote cluster, the user's local
        cloud credential files are not needed to be uploaded to the remote
        instance, as the remote instance can be assigned with a service account
        that has the necessary permissions to access the cloud resources.
        """
        return cls._SUPPORTS_SERVICE_ACCOUNT_ON_REMOTE

    #### Regions/Zones ####

    @classmethod
    def regions_with_offering(cls, instance_type: str,
                              accelerators: Optional[Dict[str, int]],
                              use_spot: bool, region: Optional[str],
                              zone: Optional[str]) -> List[Region]:
        """Returns the regions that offer the specified resources.

        The order of the regions follow the order of the regions returned by
        service_catalog/common.py#get_region_zones().
        When region or zone is not None, the returned value will be limited to
        the specified region/zone.

        Returns:
            A set of `Region`s that have the offerings for the specified
            resources.
            For each `Region` in the set, `region.zones` is the list of `Zone`s
            which have the offerings. For the clouds that do not expose `Zone`s,
            `region.zones` is an empty list.
        """
        raise NotImplementedError

    @classmethod
    def optimize_by_zone(cls) -> bool:
        """Returns whether to optimize this cloud by zone (default: region)."""
        return False

    @classmethod
    def zones_provision_loop(
        cls,
        *,
        region: str,
        num_nodes: int,
        instance_type: str,
        accelerators: Optional[Dict[str, int]] = None,
        use_spot: bool = False,
    ) -> Iterator[Optional[List[Zone]]]:
        """Loops over zones to retry for provisioning in a given region.

        Certain clouds' provisioners may handle batched requests, retrying for
        itself a list of zones under a region.  Others may need a specific zone
        per provision request (in that case, yields a one-element list for each
        zone).
        Optionally, caller can filter the yielded region/zones by specifying the
        instance_type, accelerators, and use_spot.

        Args:
            region: The region to provision.
            num_nodes: The number of nodes to provision.
            instance_type: The instance type to provision.
            accelerators: The accelerators to provision.
            use_spot: Whether to use spot instances.

        Yields:
            A list of zones that offer the requested resources in the given
            region, in the order of price.
            (1) If there is no zone that offers the specified resources, nothing
                is yielded. For example, Azure does not support zone, and
                calling this method with non-existing instance_type in the given
                region, will yield nothing, i.e. raise StopIteration.
                ```
                for zone in Azure.zones_provision_loop(region=region,
                                           instance_type='non-existing'):
                    # Will not reach here.
                ```
            (2) If the cloud's provisioner does not support `Zone`s, `None` will
                be yielded.
                ```
                for zone in Azure.zones_provision_loop(region=region,
                                           instance_type='existing-instance'):
                    assert zone is None
                ```
            This means if something is yielded, either it's None (zones are not
            supported and the region offers the resources) or it's a non-empty
            list (zones are supported and they offer the resources).

        Typical usage:

            for zones in cloud.region_zones_provision_loop(
                region,
                num_nodes,
                instance_type,
                accelerators,
                use_spot
            ):
                success = try_provision(region, zones, resources)
                if success:
                    break
        """
        raise NotImplementedError

    @classmethod
    def get_zone_shell_cmd(cls) -> Optional[str]:
        """Returns the shell command to obtain the zone of instance."""
        raise NotImplementedError

    #### Normal methods ####

    def instance_type_to_hourly_cost(self, instance_type: str, use_spot: bool,
                                     region: Optional[str],
                                     zone: Optional[str]) -> float:
        """Returns the hourly on-demand/spot price for an instance type."""
        raise NotImplementedError

    def accelerators_to_hourly_cost(self, accelerators: Dict[str, int],
                                    use_spot: bool, region: Optional[str],
                                    zone: Optional[str]) -> float:
        """Returns the hourly on-demand price for accelerators."""
        raise NotImplementedError

    def get_egress_cost(self, num_gigabytes: float):
        """Returns the egress cost.

        TODO: takes into account "per month" accumulation per account.
        """
        raise NotImplementedError

    def is_same_cloud(self, other: 'Cloud') -> bool:
        return isinstance(other, self.__class__)

    def make_deploy_resources_variables(
        self,
        resources: 'resources_lib.Resources',
        cluster_name: resources_utils.ClusterName,
        region: 'Region',
        zones: Optional[List['Zone']],
        num_nodes: int,
        dryrun: bool = False,
    ) -> Dict[str, Optional[str]]:
        """Converts planned sky.Resources to cloud-specific resource variables.

        These variables are used to fill the node type section (instance type,
        any accelerators, etc.) in the cloud's deployment YAML template.

        Cloud-agnostic sections (e.g., commands to run) need not be returned by
        this function.

        Returns:
          A dictionary of cloud-specific node type variables.
        """
        raise NotImplementedError

    @classmethod
    def get_vcpus_mem_from_instance_type(
            cls, instance_type: str) -> Tuple[Optional[float], Optional[float]]:
        """Returns the #vCPUs and memory that the instance type offers."""
        raise NotImplementedError

    @classmethod
    def get_accelerators_from_instance_type(
        cls,
        instance_type: str,
    ) -> Optional[Dict[str, Union[int, float]]]:
        """Returns {acc: acc_count} held by 'instance_type', if any."""
        raise NotImplementedError

    @classmethod
    def get_default_instance_type(
            cls,
            cpus: Optional[str] = None,
            memory: Optional[str] = None,
            disk_tier: Optional[resources_utils.DiskTier] = None
    ) -> Optional[str]:
        """Returns the default instance type with the given #vCPUs, memory and
        disk tier.

        For example, if cpus='4', this method returns the default instance type
        with 4 vCPUs.  If cpus='4+', this method returns the default instance
        type with 4 or more vCPUs.

        If 'memory=4', this method returns the default instance type with 4GB
        memory.  If 'memory=4+', this method returns the default instance
        type with 4GB or more memory.

        If disk_tier=DiskTier.MEDIUM, this method returns the default instance
        type that support medium disk tier.

        When cpus is None, memory is None or disk_tier is None, this method will
        never return None. This method may return None if the cloud's default
        instance family does not have a VM with the given number of vCPUs
        (e.g., when cpus='7') or does not have a VM with the give disk tier
        (e.g. Azure, disk_tier='high').
        """
        raise NotImplementedError

    @classmethod
    def is_image_tag_valid(cls, image_tag: str, region: Optional[str]) -> bool:
        """Validates that the image tag is valid for this cloud."""
        return service_catalog.is_image_tag_valid(image_tag,
                                                  region,
                                                  clouds=cls._REPR.lower())

    @classmethod
    def is_label_valid(cls, label_key: str,
                       label_value: str) -> Tuple[bool, Optional[str]]:
        """Validates that the label key and value are valid for this cloud.

        Labels can be implemented in different ways across clouds. For example,
        on AWS we use instance tags, on GCP we use labels, and on Kubernetes we
        use labels. This method should be implemented to validate the label
        format for the cloud.

        Returns:
            A tuple of a boolean indicating whether the label is valid and an
            optional string describing the reason if the label is invalid.
        """
        # If a cloud does not support labels, they are ignored. Only clouds
        # that support labels implement this method.
        del label_key, label_value
        return True, None

    @timeline.event
    def get_feasible_launchable_resources(
            self,
            resources: 'resources_lib.Resources',
            num_nodes: int = 1) -> 'resources_utils.FeasibleResources':
        """Returns FeasibleResources for the given resources.

        Feasible resources refer to an offering respecting the resource
        requirements.  Currently, this function implements "filtering" the
        cloud's offerings only w.r.t. accelerators constraints.

        Launchable resources require a cloud and an instance type be assigned.

        The returned dataclass object FeasibleResources contains three fields:

        - resources_list: a list of resources that are feasible to launch
        - fuzzy_candidate_list: a list of resources that loosely match requested
            resources. E.g., when A100:1 GPU is requested but is not available
            in a cloud/region, the fuzzy candidates are results of a fuzzy
            search in the catalog that are offered in the location. E.g.,
            ['A100-80GB:1', 'A100-80GB:2', 'A100-80GB:4', 'A100:8']
        - hint: an optional string hint if no feasible resources are found.
        """
        if resources.is_launchable():
            self._check_instance_type_accelerators_combination(resources)
        resources_required_features = resources.get_required_cloud_features()
        if num_nodes > 1:
            resources_required_features.add(
                CloudImplementationFeatures.MULTI_NODE)

        try:
            self.check_features_are_supported(resources,
                                              resources_required_features)
        except exceptions.NotSupportedError as e:
            # TODO(zhwu): The resources are now silently filtered out. We
            # should have some logging telling the user why the resources
            # are not considered.
            # UPDATE(kyuds): passing in NotSupportedError reason string
            # to hint for issue #5344. Did not remove above comment as
            # reason is not displayed when other resources are valid.
            return resources_utils.FeasibleResources(resources_list=[],
                                                     fuzzy_candidate_list=[],
                                                     hint=str(e))
        return self._get_feasible_launchable_resources(resources)

    def _get_feasible_launchable_resources(
        self, resources: 'resources_lib.Resources'
    ) -> 'resources_utils.FeasibleResources':
        """See get_feasible_launchable_resources()."""
        # TODO: Currently only the Kubernetes implementation of this method
        #  returns hints when no feasible resources are found. This should be
        #  implemented for all clouds.
        raise NotImplementedError

    def get_reservations_available_resources(
        self,
        instance_type: str,
        region: str,
        zone: Optional[str],
        specific_reservations: Set[str],
    ) -> Dict[str, int]:
        """"
        Returns the number of available resources per reservation for the given
        instance type in the given region/zone.
        Default implementation returns 0 for non-implemented clouds.
        """
        del instance_type, region, zone
        return {reservation: 0 for reservation in specific_reservations}

    @classmethod
    def check_credentials(
            cls,
            cloud_capability: CloudCapability) -> Tuple[bool, Optional[str]]:
        """Checks if the user has access credentials to this cloud.

        Returns a boolean of whether the user can access this cloud, and a
        string describing the reason if the user cannot access.

        Raises NotSupportedError if the capability is
        not supported by this cloud.
        """
        if cloud_capability == CloudCapability.COMPUTE:
            return cls._check_compute_credentials()
        elif cloud_capability == CloudCapability.STORAGE:
            return cls._check_storage_credentials()
        assert_never(cloud_capability)

    @classmethod
    def _check_compute_credentials(cls) -> Tuple[bool, Optional[str]]:
        """Checks if the user has access credentials to
        this cloud's compute service."""
        raise exceptions.NotSupportedError(
            f'{cls._REPR} does not support {CloudCapability.COMPUTE.value}.')

    @classmethod
    def _check_storage_credentials(cls) -> Tuple[bool, Optional[str]]:
        """Checks if the user has access credentials to
        this cloud's storage service."""
        raise exceptions.NotSupportedError(
            f'{cls._REPR} does not support {CloudCapability.STORAGE.value}.')

    # TODO(zhwu): Make the return type immutable.
    @classmethod
    def get_user_identities(cls) -> Optional[List[List[str]]]:
        """(Advanced) Returns all available user identities of this cloud.

        The user "identity" is associated with each SkyPilot cluster they
        create. This is used in protecting cluster operations, such as
        provision, teardown and status refreshing, in a multi-identity
        scenario, where the same user/device can switch between different
        cloud identities. We check that the user identity matches before:
            - Provisioning/starting a cluster
            - Stopping/tearing down a cluster
            - Refreshing the status of a cluster

        Design choices:
          1. We allow the operations that can correctly work with a different
             user identity, as a user should have full control over all their
             clusters (no matter which identity it belongs to), e.g.,
             submitting jobs, viewing logs, auto-stopping, etc.
          2. A cloud implementation can optionally switch between different
             identities if required for cluster operations. In this case,
             the cloud implementation should return multiple identities
             as a list. E.g., our Kubernetes implementation can use multiple
             kubeconfig contexts to switch between different identities.

        The choice of what constitutes an identity is up to each cloud's
        implementation. In general, to suffice for the above purposes,
        ensure that different identities should imply different sets of
        resources are used when the user invoked each cloud's default
        CLI/API.

        An identity is a list of strings. The list is in the order of
        strictness, i.e., the first element is the most strict identity, and
        the last element is the least strict identity.
        When performing an identity check between the current active identity
        and the owner identity associated with a cluster, we compare the two
        lists in order: if a position does not match, we go to the next. To
        see an example, see the docstring of the AWS.get_user_identities.

        Example identities (see cloud implementations):
            - AWS: [UserId, AccountId]
            - GCP: [email address + project ID]
            - Azure: [email address + subscription ID]
            - Kubernetes: [context name]

        Example return values:
            - AWS: [[UserId, AccountId]]
            - GCP: [[email address + project ID]]
            - Azure: [[email address + subscription ID]]
            - Kubernetes: [[current active context], [context 2], ...]

        Returns:
            None if the cloud does not have a concept of user identity
            (access protection will be disabled for these clusters);
            otherwise a list of available identities with the current active
            identity being the first element. Most clouds have only one identity
            available, so the returned list will only have one element: the
            current active identity.

        Raises:
            exceptions.CloudUserIdentityError: If the user identity cannot be
                retrieved.
        """
        return None

    @classmethod
    def get_active_user_identity_str(cls) -> Optional[str]:
        """Returns a user friendly representation of the active identity."""
        user_identity = cls.get_active_user_identity()
        if user_identity is None:
            return None
        return ', '.join(user_identity)

    @classmethod
    def get_active_user_identity(cls) -> Optional[List[str]]:
        """Returns currently active user identity of this cloud

        See get_user_identities for definition of user identity.

        Returns:
            None if the cloud does not have a concept of user identity;
            otherwise the current active identity.
        """
        identities = cls.get_user_identities()
        return identities[0] if identities is not None else None

    def get_credential_file_mounts(self) -> Dict[str, str]:
        """Returns the files necessary to access this cloud.

        Returns a dictionary that will be added to a task's file mounts.
        """
        raise NotImplementedError

    def can_credential_expire(self) -> bool:
        """Returns whether the cloud credential can expire."""
        return False

    @classmethod
    def get_image_size(cls, image_id: str, region: Optional[str]) -> float:
        """Check the image size from the cloud.

        Returns: the image size in GB.
        Raises: ValueError if the image cannot be found.
        """
        raise NotImplementedError

    def instance_type_exists(self, instance_type):
        """Returns whether the instance type exists for this cloud."""
        raise NotImplementedError

    def validate_region_zone(
            self, region: Optional[str],
            zone: Optional[str]) -> Tuple[Optional[str], Optional[str]]:
        """Validates whether region and zone exist in the catalog.

        Returns:
            A tuple of region and zone, if validated.

        Raises:
            ValueError: If region or zone is invalid or not supported.
        """
        return service_catalog.validate_region_zone(region,
                                                    zone,
                                                    clouds=self._REPR.lower())

    def need_cleanup_after_preemption_or_failure(
            self, resources: 'resources_lib.Resources') -> bool:
        """Whether a resource needs cleanup after preeemption or failure.

        In most cases, spot resources do not need cleanup after preemption,
        as long as the cluster can be relaunched with the same name and tag,
        no matter the preemption behavior is to terminate or stop the cluster.
        Similar for on-demand resources that go into maintenance mode. The
        only exception by far is GCP's TPU VM. We override this method in
        gcp.py.
        """
        del resources
        return False

    @classmethod
    def check_features_are_supported(
            cls, resources: 'resources_lib.Resources',
            requested_features: Set[CloudImplementationFeatures]) -> None:
        """Errors out if the cloud does not support all requested features.

        For instance, Lambda Cloud does not support stop, so
        Lambda.check_features_are_supported(to_provision, {
            CloudImplementationFeatures.STOP
        }) raises the exception.

        Resources are also passed as some features may depend on the resources
        requested. For example, some clouds support stopping normal instances,
        but not spot instances, e.g., AWS; or, GCP supports stopping TPU VMs but
        not TPU VM pods.

        Raises:
            exceptions.NotSupportedError: If the cloud does not support all the
            requested features.
        """
        unsupported_features2reason = cls._unsupported_features_for_resources(
            resources)

        # Docker image is not compatible with ssh proxy command.
        if skypilot_config.get_nested(
            (str(cls._REPR).lower(), 'ssh_proxy_command'), None) is not None:
            unsupported_features2reason.update({
                CloudImplementationFeatures.DOCKER_IMAGE: (
                    f'Docker image is currently not supported on {cls._REPR} '
                    'when proxy command is set. Please remove proxy command in '
                    'the config.'),
            })

        unsupported_features = set(unsupported_features2reason.keys())
        unsupported_features = requested_features.intersection(
            unsupported_features)
        if unsupported_features:
            table = log_utils.create_table(['Feature', 'Reason'])
            for feature in unsupported_features:
                table.add_row(
                    [feature.value, unsupported_features2reason[feature]])
            with ux_utils.print_exception_no_traceback():
                raise exceptions.NotSupportedError(
                    f'The following features are not supported by {cls._REPR}:'
                    '\n\t' + table.get_string().replace('\n', '\n\t'))

    @classmethod
    def _unsupported_features_for_resources(
        cls, resources: 'resources_lib.Resources'
    ) -> Dict[CloudImplementationFeatures, str]:
        """The features not supported based on the resources provided.

        This method is used by check_features_are_supported() to check if the
        cloud implementation supports all the requested features.

        Returns:
            A dict of {feature: reason} for the features not supported by the
            cloud implementation.
        """
        del resources
        raise NotImplementedError

    @classmethod
    def check_disk_tier_enabled(cls, instance_type: Optional[str],
                                disk_tier: resources_utils.DiskTier) -> None:
        """Errors out if the disk tier is not supported by the cloud provider.

        Raises:
            exceptions.NotSupportedError: If the disk tier is not supported.
        """
        del instance_type  # unused
        if disk_tier not in cls._SUPPORTED_DISK_TIERS:
            with ux_utils.print_exception_no_traceback():
                raise exceptions.NotSupportedError(
                    f'{disk_tier} is not supported by {cls._REPR}.')

    @classmethod
    def _translate_disk_tier(
        cls, disk_tier: Optional[resources_utils.DiskTier]
    ) -> resources_utils.DiskTier:
        if disk_tier is None:
            return cls._DEFAULT_DISK_TIER
        if disk_tier == resources_utils.DiskTier.BEST:
            return cls._BEST_DISK_TIER
        return disk_tier

    @classmethod
    def _check_instance_type_accelerators_combination(
            cls, resources: 'resources_lib.Resources') -> None:
        """Errors out if the accelerator is not supported by the instance type.

        This function is overridden by GCP for host-accelerator logic.

        Raises:
            ResourcesMismatchError: If the accelerator is not supported.
        """
        assert resources.is_launchable(), resources

        def _equal_accelerators(
            acc_requested: Optional[Dict[str, Union[int, float]]],
            acc_from_instance_type: Optional[Dict[str, Union[int,
                                                             float]]]) -> bool:
            """Check the requested accelerators equals to the instance type

            Check the requested accelerators equals to the accelerators
            from the instance type (both the accelerator type and the
            count).
            """
            if acc_requested is None:
                return acc_from_instance_type is None
            if acc_from_instance_type is None:
                return False

            for acc in acc_requested:
                if acc not in acc_from_instance_type:
                    return False
                # Avoid float point precision issue.
                if not math.isclose(acc_requested[acc],
                                    acc_from_instance_type[acc]):
                    return False
            return True

        acc_from_instance_type = cls.get_accelerators_from_instance_type(
            resources.instance_type)
        if not _equal_accelerators(resources.accelerators,
                                   acc_from_instance_type):
            with ux_utils.print_exception_no_traceback():
                raise exceptions.ResourcesMismatchError(
                    'Infeasible resource demands found:'
                    '\n  Instance type requested: '
                    f'{resources.instance_type}\n'
                    f'  Accelerators for {resources.instance_type}: '
                    f'{acc_from_instance_type}\n'
                    f'  Accelerators requested: {resources.accelerators}\n'
                    f'To fix: either only specify instance_type, or '
                    'change the accelerators field to be consistent.')

    @classmethod
    def check_quota_available(cls,
                              resources: 'resources_lib.Resources') -> bool:
        """Check if quota is available based on `resources`.

        The _retry_zones function in cloud_vm_ray_backend goes through different
        candidate regions and attempts to provision the requested
        `instance_type` or `accelerator` accelerators in the `region`
        (the `instance_type` or `accelerator`, and `region`, as defined in
        `resources`) until a successful provisioning happens or all regions
        with the requested accelerator have been looked at. Previously,
        SkyPilot would attempt to provision resources in all of these regions.
        However, many regions would have a zero quota or inadequate quota,
        meaning these attempted provisions were destined to fail from
        the get-go.

        Checking the quota is substantially faster than attempting a failed
        provision (~1 second vs 30+ seconds) so this function attempts to
        check the resource quota and return False if it is found to be zero,
        or True otherwise. If False is returned, _retry_zones will not attempt
        a provision in the region, saving time.

        We are only checking for a nonzero quota, instead of also factoring in
        quota utilization because many cloud providers' APIs don't have a
        built-in command for checking the real-time utilization. Checking
        real-time utilization is a more difficult endeavor that involves
        observability etc., so we are holding off on that for now.

        If for at any point the function fails, whether it's because we can't
        import the necessary dependencies or a query using a cloud provider's
        API fails, we will return True, because we cannot conclusively say the
        relevant quota is zero in these cases, and we don't want to
        preemptively exclude regions from an attempted provision if they may
        have an adequate quota.

        Design choice: We chose a just-in-time approach where
        check_quota_available is called immediately before a potential
        attempted provision, rather than checking all region quotas
        beforehand, storing them, and using those values on-demand. This is
        because, for example, _retry_zones may only need to go through one or
        a few regions before a successful region, and running a query to check
        *every* region's quota beforehand would cause an unnecessary delay.

        Returns:
            False if the quota is found to be zero, and true otherwise.
        """
        del resources  # unused

        return True

    @classmethod
    def query_status(cls, name: str, tag_filters: Dict[str, str],
                     region: Optional[str], zone: Optional[str],
                     **kwargs) -> List['status_lib.ClusterStatus']:
        """Queries the latest status of the cluster from the cloud.

        The global_user_state caches the status of the clusters, but the
        actual status of the clusters may change on the cloud, e.g., the
        autostop happens, or the user manually stops the cluster. This
        method queries the cloud to get the latest cluster status.

        Returns:
            A list of ClusterStatus representing the status of all the
            alive nodes in the cluster.

        Raises:
            exceptions.ClusterStatusFetchingError: raised if the status of the
                cluster cannot be fetched.
        """
        raise NotImplementedError

    # === Image related methods ===
    # These three methods are used to create, move and delete images. They
    # are currently only used in `sky launch --clone-disk-from` to clone a
    # cluster's disk to launch a new cluster.
    # It is not required to implement these methods for clouds that do not
    # support `--clone-disk-from`. If not implemented,
    # CloudImplementationFeatures.CLONE_DISK_FROM should be added to the
    # cloud._cloud_unsupported_features().

    @classmethod
    def create_image_from_cluster(cls,
                                  cluster_name: resources_utils.ClusterName,
                                  region: Optional[str],
                                  zone: Optional[str]) -> str:
        """Creates an image from the cluster.

        Returns: the image ID.
        """
        raise NotImplementedError

    @classmethod
    def maybe_move_image(cls, image_id: str, source_region: str,
                         target_region: str, source_zone: Optional[str],
                         target_zone: Optional[str]) -> str:
        """Move an image if required.

        If the image cannot be accessed in the target region, move the image
        from the source region to the target region.

        Returns: the image ID in the target region.
        """
        raise NotImplementedError

    @classmethod
    def delete_image(cls, image_id: str, region: Optional[str]) -> None:
        """Deletes the image with image_id in the region."""
        raise NotImplementedError

    # === End of image related methods ===

    @classmethod
    def canonical_name(cls) -> str:
        return cls.__name__.lower()

    def __repr__(self):
        return self._REPR

    def __getstate__(self):
        state = self.__dict__.copy()
        state.pop('PROVISIONER_VERSION', None)
        state.pop('STATUS_VERSION', None)
        return state


# === Helper functions ===
def cloud_in_iterable(cloud: Cloud, cloud_list: Iterable[Cloud]) -> bool:
    """Returns whether the cloud is in the given cloud list."""
    return any(cloud.is_same_cloud(c) for c in cloud_list)



"""Logging utilities."""
import builtins
import contextlib
from datetime import datetime
import logging
import os
import sys
import threading

import colorama

from sky.skylet import constants
from sky.utils import env_options
from sky.utils import rich_utils

# UX: Should we show logging prefixes and some extra information in optimizer?
_FORMAT = '%(levelname).1s %(asctime)s %(filename)s:%(lineno)d] %(message)s'
_DATE_FORMAT = '%m-%d %H:%M:%S'
_SENSITIVE_LOGGER = ['sky.provisioner', 'sky.optimizer']

DEBUG = logging.DEBUG
INFO = logging.INFO
WARNING = logging.WARNING
ERROR = logging.ERROR
CRITICAL = logging.CRITICAL


def _show_logging_prefix():
    return env_options.Options.SHOW_DEBUG_INFO.get(
    ) or not env_options.Options.MINIMIZE_LOGGING.get()


class NewLineFormatter(logging.Formatter):
    """Adds logging prefix to newlines to align multi-line messages."""

    def __init__(self, fmt, datefmt=None, dim=False):
        logging.Formatter.__init__(self, fmt, datefmt)
        self.dim = dim

    def format(self, record):
        msg = logging.Formatter.format(self, record)
        if record.message != '':
            parts = msg.partition(record.message)
            msg = msg.replace('\n', '\r\n' + parts[0])
            if self.dim:
                msg = colorama.Style.DIM + msg + colorama.Style.RESET_ALL
        return msg


_root_logger = logging.getLogger('sky')
_default_handler = None
_logging_config = threading.local()

NO_PREFIX_FORMATTER = NewLineFormatter(None, datefmt=_DATE_FORMAT)
FORMATTER = NewLineFormatter(_FORMAT, datefmt=_DATE_FORMAT)
DIM_FORMATTER = NewLineFormatter(_FORMAT, datefmt=_DATE_FORMAT, dim=True)

# All code inside the library should use sky_logging.print()
# rather than print().
# This is to make controlled logging via is_silent() possible:
# in some situation we would like to disable any
# printing/logging.
print = builtins.print  # pylint: disable=redefined-builtin


def _setup_logger():
    _root_logger.setLevel(logging.DEBUG)
    global _default_handler
    if _default_handler is None:
        _default_handler = rich_utils.RichSafeStreamHandler(sys.stdout)
        _default_handler.flush = sys.stdout.flush  # type: ignore
        if env_options.Options.SHOW_DEBUG_INFO.get():
            _default_handler.setLevel(logging.DEBUG)
        else:
            _default_handler.setLevel(logging.INFO)
        _root_logger.addHandler(_default_handler)
    if _show_logging_prefix():
        _default_handler.setFormatter(FORMATTER)
    else:
        _default_handler.setFormatter(NO_PREFIX_FORMATTER)
    # Setting this will avoid the message
    # being propagated to the parent logger.
    _root_logger.propagate = False
    if env_options.Options.SUPPRESS_SENSITIVE_LOG.get():
        # If the sensitive log is enabled, we reinitialize a new handler
        # and force set the level to INFO to suppress the debug logs
        # for certain loggers.
        for logger_name in _SENSITIVE_LOGGER:
            logger = logging.getLogger(logger_name)
            handler_to_logger = rich_utils.RichSafeStreamHandler(sys.stdout)
            handler_to_logger.flush = sys.stdout.flush  # type: ignore
            logger.addHandler(handler_to_logger)
            logger.setLevel(logging.INFO)
            if _show_logging_prefix():
                handler_to_logger.setFormatter(FORMATTER)
            else:
                handler_to_logger.setFormatter(NO_PREFIX_FORMATTER)
            # Do not propagate to the parent logger to avoid parent
            # logger printing the logs.
            logger.propagate = False


def reload_logger():
    """Reload the logger.

    This ensures that the logger takes the new environment variables,
    such as SKYPILOT_DEBUG.
    """
    global _default_handler
    _root_logger.removeHandler(_default_handler)
    _default_handler = None
    _setup_logger()


# The logger is initialized when the module is imported.
# This is thread-safe as the module is only imported once,
# guaranteed by the Python GIL.
_setup_logger()


def init_logger(name: str) -> logging.Logger:
    return logging.getLogger(name)


@contextlib.contextmanager
def set_logging_level(logger: str, level: int):
    logger = logging.getLogger(logger)
    original_level = logger.level
    logger.setLevel(level)
    try:
        yield
    finally:
        logger.setLevel(original_level)


def logging_enabled(logger: logging.Logger, level: int) -> bool:
    return logger.level <= level


@contextlib.contextmanager
def silent():
    """Make all sky_logging.print() and logger.{info, warning...} silent.

    We preserve the ERROR level logging, so that errors are
    still printed.
    """
    global print
    previous_level = _root_logger.level
    previous_is_silent = is_silent()
    previous_print = print

    # Turn off logger
    _root_logger.setLevel(logging.ERROR)
    _logging_config.is_silent = True
    print = lambda *args, **kwargs: None
    try:
        yield
    finally:
        # Restore logger
        print = previous_print
        _root_logger.setLevel(previous_level)
        _logging_config.is_silent = previous_is_silent


def is_silent():
    if not hasattr(_logging_config, 'is_silent'):
        # Should not set it globally, as the global assignment
        # will be executed only once if the module is imported
        # in the main thread, and will not be executed in other
        # threads.
        _logging_config.is_silent = False
    return _logging_config.is_silent


def get_run_timestamp() -> str:
    return 'sky-' + datetime.now().strftime('%Y-%m-%d-%H-%M-%S-%f')


def generate_tmp_logging_file_path(file_name: str) -> str:
    """Generate an absolute path of a tmp file for logging."""
    run_timestamp = get_run_timestamp()
    log_dir = os.path.join(constants.SKY_LOGS_DIRECTORY, run_timestamp)
    log_path = os.path.expanduser(os.path.join(log_dir, file_name))

    return log_path



"""Utils for AWS provisioner."""
import colorama

from sky import exceptions
from sky import sky_logging

logger = sky_logging.init_logger(__name__)

BOTO_MAX_RETRIES = 12

# ======================== Thread-safe ========================
# https://boto3.amazonaws.com/v1/documentation/api/latest/guide/resources.html#multithreading-or-multiprocessing-with-resources


def handle_boto_error(exc: Exception, msg: str) -> None:
    """Handle boto3 error properly."""
    error_code = None
    error_info = None
    # todo: not sure if these exceptions always have response
    if hasattr(exc, 'response'):
        error_info = exc.response.get('Error', None)
    if error_info is not None:
        error_code = error_info.get('Code', None)

    generic_message = (f'{msg}\nError code: {colorama.Style.BRIGHT}{error_code}'
                       f'{colorama.Style.RESET_ALL}')

    # For invalid credentials, like expired tokens or
    # invalid tokens, raise InvalidCloudCredentials exception
    invalid_credentials_codes = [
        'ExpiredTokenException',
        'ExpiredToken',
        'RequestExpired',
        'InvalidClientTokenId',
        'InvalidClientToken',
    ]

    if error_code in invalid_credentials_codes:
        # 'An error occurred (ExpiredToken) when calling the
        # GetInstanceProfile operation: The security token
        # included in the request is expired'

        # 'An error occurred (RequestExpired) when calling the
        # DescribeKeyPairs operation: Request has expired.'

        token_command = ('aws sts get-session-token '
                         '--serial-number arn:aws:iam::' + 'ROOT_ACCOUNT_ID' +
                         ':mfa/' + 'AWS_USERNAME' + ' --token-code ' +
                         'TWO_FACTOR_AUTH_CODE')

        secret_key_var = ('export AWS_SECRET_ACCESS_KEY = ' + 'REPLACE_ME' +
                          ' # found at Credentials.SecretAccessKey')
        session_token_var = ('export AWS_SESSION_TOKEN = ' + 'REPLACE_ME' +
                             ' # found at Credentials.SessionToken')
        access_key_id_var = ('export AWS_ACCESS_KEY_ID = ' + 'REPLACE_ME' +
                             ' # found at Credentials.AccessKeyId')

        # fixme: replace with a Github URL that points
        # to our repo
        aws_session_script_url = (
            'https://gist.github.com/maximsmol/a0284e1d97b25d417bd9ae02e5f450cf'
        )

        logger.error(generic_message)
        logger.info(vars(exc))

        logger.fatal(
            'Your AWS session has expired.\n'
            f'You can request a new one using {colorama.Style.BRIGHT}'
            f'{token_command}{colorama.Style.RESET_ALL} then expose it '
            f'to SkyPilot by setting {colorama.Style.BRIGHT}'
            f'{secret_key_var} {session_token_var} {access_key_id_var}'
            f'{colorama.Style.RESET_ALL}\n'
            f'You can find a script that automates this at:'
            f'{aws_session_script_url}')
        # Raise the InvalidCloudCredentials exception so that
        # the provisioner can failover to other clouds
        raise exceptions.InvalidCloudCredentials(
            f'InvalidCloudCredentials: {generic_message}') from exc

    # todo: any other errors that we should catch separately?

    logger.fatal(generic_message)
    logger.fatal('')
    logger.error(f'Boto3 error: {vars(exc)} {exec}')
    raise SystemExit(1)



"""SDK functions for managed jobs."""
import json
import typing
from typing import Dict, List, Optional, Union
import webbrowser

import click

from sky import sky_logging
from sky.adaptors import common as adaptors_common
from sky.client import common as client_common
from sky.client import sdk
from sky.server import common as server_common
from sky.server.requests import payloads
from sky.skylet import constants
from sky.usage import usage_lib
from sky.utils import common_utils
from sky.utils import dag_utils

if typing.TYPE_CHECKING:
    import io

    import requests

    import sky
else:
    requests = adaptors_common.LazyImport('requests')

logger = sky_logging.init_logger(__name__)


@usage_lib.entrypoint
@server_common.check_server_healthy_or_start
def launch(
    task: Union['sky.Task', 'sky.Dag'],
    name: Optional[str] = None,
    # Internal only:
    # pylint: disable=invalid-name
    _need_confirmation: bool = False,
) -> server_common.RequestId:
    """Launches a managed job.

    Please refer to sky.cli.job_launch for documentation.

    Args:
        task: sky.Task, or sky.Dag (experimental; 1-task only) to launch as a
            managed job.
        name: Name of the managed job.
        _need_confirmation: (Internal only) Whether to show a confirmation
            prompt before launching the job.

    Returns:
        The request ID of the launch request.

    Request Returns:
        job_id (Optional[int]): Job ID for the managed job
        controller_handle (Optional[ResourceHandle]): ResourceHandle of the
          controller

    Request Raises:
        ValueError: cluster does not exist. Or, the entrypoint is not a valid
          chain dag.
        sky.exceptions.NotSupportedError: the feature is not supported.
    """

    dag = dag_utils.convert_entrypoint_to_dag(task)
    sdk.validate(dag)
    if _need_confirmation:
        request_id = sdk.optimize(dag)
        sdk.stream_and_get(request_id)
        prompt = f'Launching a managed job {dag.name!r}. Proceed?'
        if prompt is not None:
            click.confirm(prompt, default=True, abort=True, show_default=True)

    dag = client_common.upload_mounts_to_api_server(dag)
    dag_str = dag_utils.dump_chain_dag_to_yaml_str(dag)
    body = payloads.JobsLaunchBody(
        task=dag_str,
        name=name,
    )
    response = requests.post(
        f'{server_common.get_server_url()}/jobs/launch',
        json=json.loads(body.model_dump_json()),
        timeout=(5, None),
        cookies=server_common.get_api_cookie_jar(),
    )
    return server_common.get_request_id(response)


@usage_lib.entrypoint
@server_common.check_server_healthy_or_start
def queue(refresh: bool,
          skip_finished: bool = False,
          all_users: bool = False) -> server_common.RequestId:
    """Gets statuses of managed jobs.

    Please refer to sky.cli.job_queue for documentation.

    Args:
        refresh: Whether to restart the jobs controller if it is stopped.
        skip_finished: Whether to skip finished jobs.
        all_users: Whether to show all users' jobs.

    Returns:
        The request ID of the queue request.

    Request Returns:
        job_records (List[Dict[str, Any]]): A list of dicts, with each dict
          containing the information of a job.

          .. code-block:: python

            [
              {
                'job_id': (int) job id,
                'job_name': (str) job name,
                'resources': (str) resources of the job,
                'submitted_at': (float) timestamp of submission,
                'end_at': (float) timestamp of end,
                'duration': (float) duration in seconds,
                'recovery_count': (int) Number of retries,
                'status': (sky.jobs.ManagedJobStatus) of the job,
                'cluster_resources': (str) resources of the cluster,
                'region': (str) region of the cluster,
              }
            ]

    Request Raises:
        sky.exceptions.ClusterNotUpError: the jobs controller is not up or
          does not exist.
        RuntimeError: if failed to get the managed jobs with ssh.
    """
    body = payloads.JobsQueueBody(
        refresh=refresh,
        skip_finished=skip_finished,
        all_users=all_users,
    )
    response = requests.post(
        f'{server_common.get_server_url()}/jobs/queue',
        json=json.loads(body.model_dump_json()),
        timeout=(5, None),
        cookies=server_common.get_api_cookie_jar(),
    )
    return server_common.get_request_id(response=response)


@usage_lib.entrypoint
@server_common.check_server_healthy_or_start
def cancel(
    name: Optional[str] = None,
    job_ids: Optional[List[int]] = None,
    all: bool = False,  # pylint: disable=redefined-builtin
    all_users: bool = False,
) -> server_common.RequestId:
    """Cancels managed jobs.

    Please refer to sky.cli.job_cancel for documentation.

    Args:
        name: Name of the managed job to cancel.
        job_ids: IDs of the managed jobs to cancel.
        all: Whether to cancel all managed jobs.
        all_users: Whether to cancel all managed jobs from all users.

    Returns:
        The request ID of the cancel request.

    Request Raises:
        sky.exceptions.ClusterNotUpError: the jobs controller is not up.
        RuntimeError: failed to cancel the job.
    """
    body = payloads.JobsCancelBody(
        name=name,
        job_ids=job_ids,
        all=all,
        all_users=all_users,
    )
    response = requests.post(
        f'{server_common.get_server_url()}/jobs/cancel',
        json=json.loads(body.model_dump_json()),
        timeout=(5, None),
        cookies=server_common.get_api_cookie_jar(),
    )
    return server_common.get_request_id(response=response)


@usage_lib.entrypoint
@server_common.check_server_healthy_or_start
def tail_logs(name: Optional[str] = None,
              job_id: Optional[int] = None,
              follow: bool = True,
              controller: bool = False,
              refresh: bool = False,
              output_stream: Optional['io.TextIOBase'] = None) -> int:
    """Tails logs of managed jobs.

    You can provide either a job name or a job ID to tail logs. If both are not
    provided, the logs of the latest job will be shown.

    Args:
        name: Name of the managed job to tail logs.
        job_id: ID of the managed job to tail logs.
        follow: Whether to follow the logs.
        controller: Whether to tail logs from the jobs controller.
        refresh: Whether to restart the jobs controller if it is stopped.
        output_stream: The stream to write the logs to. If None, print to the
            console.

    Returns:
        Exit code based on success or failure of the job. 0 if success,
        100 if the job failed. See exceptions.JobExitCode for possible exit
        codes.

    Request Raises:
        ValueError: invalid arguments.
        sky.exceptions.ClusterNotUpError: the jobs controller is not up.
    """
    body = payloads.JobsLogsBody(
        name=name,
        job_id=job_id,
        follow=follow,
        controller=controller,
        refresh=refresh,
    )
    response = requests.post(
        f'{server_common.get_server_url()}/jobs/logs',
        json=json.loads(body.model_dump_json()),
        stream=True,
        timeout=(5, None),
        cookies=server_common.get_api_cookie_jar(),
    )
    request_id = server_common.get_request_id(response)
    return sdk.stream_response(request_id, response, output_stream)


@usage_lib.entrypoint
@server_common.check_server_healthy_or_start
def download_logs(
        name: Optional[str],
        job_id: Optional[int],
        refresh: bool,
        controller: bool,
        local_dir: str = constants.SKY_LOGS_DIRECTORY) -> Dict[int, str]:
    """Sync down logs of managed jobs.

    Please refer to sky.cli.job_logs for documentation.

    Args:
        name: Name of the managed job to sync down logs.
        job_id: ID of the managed job to sync down logs.
        refresh: Whether to restart the jobs controller if it is stopped.
        controller: Whether to sync down logs from the jobs controller.
        local_dir: Local directory to sync down logs.

    Returns:
        A dictionary mapping job ID to the local path.

    Request Raises:
        ValueError: invalid arguments.
        sky.exceptions.ClusterNotUpError: the jobs controller is not up.
    """

    body = payloads.JobsDownloadLogsBody(
        name=name,
        job_id=job_id,
        refresh=refresh,
        controller=controller,
        local_dir=local_dir,
    )
    response = requests.post(
        f'{server_common.get_server_url()}/jobs/download_logs',
        json=json.loads(body.model_dump_json()),
        timeout=(5, None),
        cookies=server_common.get_api_cookie_jar(),
    )
    job_id_remote_path_dict = sdk.stream_and_get(
        server_common.get_request_id(response))
    remote2local_path_dict = client_common.download_logs_from_api_server(
        job_id_remote_path_dict.values())
    return {
        job_id: remote2local_path_dict[remote_path]
        for job_id, remote_path in job_id_remote_path_dict.items()
    }


spot_launch = common_utils.deprecated_function(
    launch,
    name='sky.jobs.launch',
    deprecated_name='spot_launch',
    removing_version='0.8.0',
    override_argument={'use_spot': True})
spot_queue = common_utils.deprecated_function(queue,
                                              name='sky.jobs.queue',
                                              deprecated_name='spot_queue',
                                              removing_version='0.8.0')
spot_cancel = common_utils.deprecated_function(cancel,
                                               name='sky.jobs.cancel',
                                               deprecated_name='spot_cancel',
                                               removing_version='0.8.0')
spot_tail_logs = common_utils.deprecated_function(
    tail_logs,
    name='sky.jobs.tail_logs',
    deprecated_name='spot_tail_logs',
    removing_version='0.8.0')


@usage_lib.entrypoint
@server_common.check_server_healthy_or_start
def dashboard() -> None:
    """Starts a dashboard for managed jobs."""
    user_hash = common_utils.get_user_hash()
    api_server_url = server_common.get_server_url()
    params = f'user_hash={user_hash}'
    url = f'{api_server_url}/jobs/dashboard?{params}'
    logger.info(f'Opening dashboard in browser: {url}')
    webbrowser.open(url)



"""Skylet configs."""
import functools
import os
import pathlib
from typing import Callable, Optional, Union

from sky.utils import db_utils

_DB_PATH = os.path.expanduser('~/.sky/skylet_config.db')
os.makedirs(pathlib.Path(_DB_PATH).parents[0], exist_ok=True)

_table_created = False


def ensure_table(func: Callable):
    """Ensure the table exists before calling the function.

    Since this module will be imported whenever `sky` is imported (due to
    Python's package importing logic), we should avoid creating the table
    until it's actually needed to avoid too many concurrent commit to the
    database.
    It solves the database locked problem in #1576.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        global _table_created
        if not _table_created:
            with db_utils.safe_cursor(
                    _DB_PATH) as c:  # Call it 'c' to avoid pylint complaining.
                # Use WAL mode to avoid locking problem in #1507.
                # Reference: https://stackoverflow.com/a/39265148
                c.execute('PRAGMA journal_mode=WAL')
                c.execute("""\
                    CREATE TABLE IF NOT EXISTS config (
                        key TEXT PRIMARY KEY,
                        value TEXT)""")
        _table_created = True
        return func(*args, **kwargs)

    return wrapper


@ensure_table
def get_config(key: str) -> Optional[bytes]:
    with db_utils.safe_cursor(_DB_PATH) as cursor:
        rows = cursor.execute('SELECT value FROM config WHERE key = ?', (key,))
        for (value,) in rows:
            return value
        return None


@ensure_table
def set_config(key: str, value: Union[bytes, str]) -> None:
    with db_utils.safe_cursor(_DB_PATH) as cursor:
        cursor.execute(
            """\
            INSERT OR REPLACE INTO config VALUES (?, ?)
            """, (key, value))



"""Execution layer.

See `Stage` for a Task's life cycle.
"""
import enum
import typing
from typing import List, Optional, Tuple, Union

import colorama

from sky import admin_policy
from sky import backends
from sky import clouds
from sky import global_user_state
from sky import optimizer
from sky import sky_logging
from sky.backends import backend_utils
from sky.usage import usage_lib
from sky.utils import admin_policy_utils
from sky.utils import common
from sky.utils import controller_utils
from sky.utils import dag_utils
from sky.utils import resources_utils
from sky.utils import rich_utils
from sky.utils import status_lib
from sky.utils import timeline
from sky.utils import ux_utils

if typing.TYPE_CHECKING:
    import sky

logger = sky_logging.init_logger(__name__)


class Stage(enum.Enum):
    """Stages for a run of a sky.Task."""
    # TODO: rename actual methods to be consistent.
    CLONE_DISK = enum.auto()
    OPTIMIZE = enum.auto()
    PROVISION = enum.auto()
    SYNC_WORKDIR = enum.auto()
    SYNC_FILE_MOUNTS = enum.auto()
    SETUP = enum.auto()
    PRE_EXEC = enum.auto()
    EXEC = enum.auto()
    DOWN = enum.auto()


def _maybe_clone_disk_from_cluster(clone_disk_from: Optional[str],
                                   cluster_name: Optional[str],
                                   task: 'sky.Task') -> 'sky.Task':
    if clone_disk_from is None:
        return task
    task, handle = backend_utils.check_can_clone_disk_and_override_task(
        clone_disk_from, cluster_name, task)
    original_cloud = handle.launched_resources.cloud
    assert original_cloud is not None, handle.launched_resources
    task_resources = list(task.resources)[0]

    with rich_utils.safe_status('Creating image from source cluster '
                                f'{clone_disk_from!r}'):
        image_id = original_cloud.create_image_from_cluster(
            cluster_name=resources_utils.ClusterName(
                display_name=clone_disk_from,
                name_on_cloud=handle.cluster_name_on_cloud),
            region=handle.launched_resources.region,
            zone=handle.launched_resources.zone,
        )
        rich_utils.force_update_status(
            f'Migrating image {image_id} to target region '
            f'{task_resources.region}...')
        source_region = handle.launched_resources.region
        target_region = task_resources.region
        assert source_region is not None, handle.launched_resources
        assert target_region is not None, task_resources

        image_id = original_cloud.maybe_move_image(
            image_id,
            source_region=source_region,
            target_region=target_region,
            source_zone=handle.launched_resources.zone,
            target_zone=task_resources.zone,
        )
    logger.info(
        f'{colorama.Fore.GREEN}'
        f'Successfully created image {image_id!r} for {clone_disk_from!r} '
        f'on {original_cloud}.{colorama.Style.RESET_ALL}\n'
        'Overriding task\'s image_id.')
    task_resources = task_resources.copy(image_id=image_id,
                                         _is_image_managed=True)
    task.set_resources(task_resources)
    # Set the best_resources to None to trigger a re-optimization, so that
    # the new task_resources is used.
    task.best_resources = None
    logger.debug(f'Overridden task resources: {task.resources}')
    return task


def _execute(
    entrypoint: Union['sky.Task', 'sky.Dag'],
    dryrun: bool = False,
    down: bool = False,
    stream_logs: bool = True,
    handle: Optional[backends.ResourceHandle] = None,
    backend: Optional[backends.Backend] = None,
    retry_until_up: bool = False,
    optimize_target: common.OptimizeTarget = common.OptimizeTarget.COST,
    stages: Optional[List[Stage]] = None,
    cluster_name: Optional[str] = None,
    detach_setup: bool = False,
    detach_run: bool = False,
    idle_minutes_to_autostop: Optional[int] = None,
    no_setup: bool = False,
    clone_disk_from: Optional[str] = None,
    skip_unnecessary_provisioning: bool = False,
    # Internal only:
    # pylint: disable=invalid-name
    _quiet_optimizer: bool = False,
    _is_launched_by_jobs_controller: bool = False,
    _is_launched_by_sky_serve_controller: bool = False,
) -> Tuple[Optional[int], Optional[backends.ResourceHandle]]:
    """Execute an entrypoint.

    If sky.Task is given or DAG has not been optimized yet, this will call
    sky.optimize() for the caller.

    Args:
      entrypoint: sky.Task or sky.Dag.
      dryrun: bool; if True, only print the provision info (e.g., cluster
        yaml).
      down: bool; whether to tear down the launched resources after all jobs
        finish (successfully or abnormally). If idle_minutes_to_autostop is
        also set, the cluster will be torn down after the specified idle time.
        Note that if errors occur during provisioning/data syncing/setting up,
        the cluster will not be torn down for debugging purposes.
      stream_logs: bool; whether to stream all tasks' outputs to the client.
      handle: Optional[backends.ResourceHandle]; if provided, execution will
        attempt to use an existing backend cluster handle instead of
        provisioning a new one.
      backend: Backend; backend to use for executing the tasks. Defaults to
        CloudVmRayBackend()
      retry_until_up: bool; whether to retry the provisioning until the cluster
        is up.
      optimize_target: OptimizeTarget; the dag optimization metric, e.g.
        OptimizeTarget.COST.
      stages: List of stages to run.  If None, run the whole life cycle of
        execution; otherwise, just the specified stages.  Used for `sky exec`
        skipping all setup steps.
      cluster_name: Name of the cluster to create/reuse.  If None,
        auto-generate a name.
      detach_setup: If True, run setup in non-interactive mode as part of the
        job itself. You can safely ctrl-c to detach from logging, and it will
        not interrupt the setup process. To see the logs again after detaching,
        use `sky logs`. To cancel setup, cancel the job via `sky cancel`.
      detach_run: If True, as soon as a job is submitted, return from this
        function and do not stream execution logs.
      idle_minutes_to_autostop: int; if provided, the cluster will be set to
        autostop after this many minutes of idleness.
      no_setup: bool; whether to skip setup commands or not when (re-)launching.
      clone_disk_from: Optional[str]; if set, clone the disk from the specified
        cluster.
      skip_unnecessary_provisioning: bool; if True, compare the calculated
        cluster config to the current cluster's config. If they match, shortcut
        provisioning and setup, even if we have Stage.PROVISION and Stage.SETUP.

    Returns:
      job_id: Optional[int]; the job ID of the submitted job. None if the
        backend is not CloudVmRayBackend, or no job is submitted to
        the cluster.
      handle: Optional[backends.ResourceHandle]; the handle to the cluster. None
        if dryrun.
    """

    dag = dag_utils.convert_entrypoint_to_dag(entrypoint)
    for task in dag.tasks:
        if task.storage_mounts is not None:
            for storage in task.storage_mounts.values():
                # Ensure the storage is constructed.
                storage.construct()
    dag, _ = admin_policy_utils.apply(
        dag,
        request_options=admin_policy.RequestOptions(
            cluster_name=cluster_name,
            idle_minutes_to_autostop=idle_minutes_to_autostop,
            down=down,
            dryrun=dryrun,
        ))
    assert len(dag) == 1, f'We support 1 task for now. {dag}'
    task = dag.tasks[0]

    if any(r.job_recovery is not None for r in task.resources):
        logger.warning(
            f'{colorama.Style.DIM}The task has `job_recovery` specified, '
            'but is launched as an unmanaged job. It will be ignored.'
            'To enable job recovery, use managed jobs: sky jobs launch.'
            f'{colorama.Style.RESET_ALL}')

    cluster_exists = False
    if cluster_name is not None:
        cluster_record = global_user_state.get_cluster_from_name(cluster_name)
        cluster_exists = cluster_record is not None
        # TODO(woosuk): If the cluster exists, print a warning that
        # `cpus` and `memory` are not used as a job scheduling constraint,
        # unlike `gpus`.

    stages = stages if stages is not None else list(Stage)

    # Requested features that some clouds support and others don't.
    requested_features = set()

    is_controller_high_availability_supported = False

    controller = controller_utils.Controllers.from_name(cluster_name)
    if controller is not None:
        requested_features.add(
            clouds.CloudImplementationFeatures.HOST_CONTROLLERS)
        if controller_utils.high_availability_specified(cluster_name,
                                                        skip_warning=False):
            requested_features.add(clouds.CloudImplementationFeatures.
                                   HIGH_AVAILABILITY_CONTROLLERS)
            # If we provision a cluster that supports high availability
            # controllers, we can use the high availability controller.
            is_controller_high_availability_supported = True

    # Add requested features from the task
    requested_features |= task.get_required_cloud_features()

    backend = backend if backend is not None else backends.CloudVmRayBackend()
    if isinstance(backend, backends.CloudVmRayBackend):
        if down and idle_minutes_to_autostop is None:
            # Use auto{stop,down} to terminate the cluster after the task is
            # done.
            idle_minutes_to_autostop = 0
        if idle_minutes_to_autostop is not None:
            if idle_minutes_to_autostop == 0:
                # idle_minutes_to_autostop=0 can cause the following problem:
                # After we set the autostop in the PRE_EXEC stage with -i 0,
                # it could be possible that the cluster immediately found
                # itself have no task running and start the auto{stop,down}
                # process, before the task is submitted in the EXEC stage.
                verb = 'torn down' if down else 'stopped'
                logger.info(f'{colorama.Style.DIM}The cluster will '
                            f'be {verb} after 1 minutes of idleness '
                            '(after all jobs finish).'
                            f'{colorama.Style.RESET_ALL}')
                idle_minutes_to_autostop = 1
            if Stage.DOWN in stages:
                stages.remove(Stage.DOWN)
            if idle_minutes_to_autostop >= 0:
                if down:
                    requested_features.add(
                        clouds.CloudImplementationFeatures.AUTODOWN)
                else:
                    requested_features.add(
                        clouds.CloudImplementationFeatures.AUTOSTOP)
        # NOTE: in general we may not have sufficiently specified info
        # (cloud/resource) to check STOP_SPOT_INSTANCE here. This is checked in
        # the backend.

    elif idle_minutes_to_autostop is not None:
        # TODO(zhwu): Autostop is not supported for non-CloudVmRayBackend.
        with ux_utils.print_exception_no_traceback():
            raise ValueError(
                f'Backend {backend.NAME} does not support autostop, please try'
                f' {backends.CloudVmRayBackend.NAME}')

    if Stage.CLONE_DISK in stages:
        task = _maybe_clone_disk_from_cluster(clone_disk_from, cluster_name,
                                              task)

    if not cluster_exists:
        # If spot is launched on serve or jobs controller, we don't need to
        # print out the hint.
        if (Stage.PROVISION in stages and task.use_spot and
                not _is_launched_by_jobs_controller and
                not _is_launched_by_sky_serve_controller):
            yellow = colorama.Fore.YELLOW
            bold = colorama.Style.BRIGHT
            reset = colorama.Style.RESET_ALL
            logger.info(
                f'{yellow}Launching a spot job that does not '
                f'automatically recover from preemptions. To '
                'get automatic recovery, use managed job instead: '
                f'{reset}{bold}sky jobs launch{reset} {yellow}or{reset} '
                f'{bold}sky.jobs.launch(){reset}.')

        if Stage.OPTIMIZE in stages:
            if task.best_resources is None:
                # TODO: fix this for the situation where number of requested
                # accelerators is not an integer.
                if isinstance(backend, backends.CloudVmRayBackend):
                    # TODO: adding this check because docker backend on a
                    # no-credential machine should not enter optimize(), which
                    # would directly error out ('No cloud is enabled...').  Fix
                    # by moving `sky check` checks out of optimize()?
                    controller = controller_utils.Controllers.from_name(
                        cluster_name)
                    if controller is not None:
                        logger.info(
                            f'Choosing resources for {controller.value.name}...'
                        )
                    dag = optimizer.Optimizer.optimize(dag,
                                                       minimize=optimize_target,
                                                       quiet=_quiet_optimizer)
                    task = dag.tasks[0]  # Keep: dag may have been deep-copied.
                    assert task.best_resources is not None, task

    backend.register_info(
        dag=dag,
        optimize_target=optimize_target,
        requested_features=requested_features,
        # That's because we want to do commands in task.setup and task.run again
        # after K8S pod recovers from a crash.
        # See `kubernetes-ray.yml.j2` for more details.
        dump_final_script=is_controller_high_availability_supported)

    if task.storage_mounts is not None:
        # Optimizer should eventually choose where to store bucket
        task.sync_storage_mounts()

    try:
        provisioning_skipped = False
        if Stage.PROVISION in stages:
            assert handle is None or skip_unnecessary_provisioning, (
                'Provisioning requested, but handle is already set. PROVISION '
                'should be excluded from stages or '
                'skip_unecessary_provisioning should be set. ')
            (handle, provisioning_skipped) = backend.provision(
                task,
                task.best_resources,
                dryrun=dryrun,
                stream_logs=stream_logs,
                cluster_name=cluster_name,
                retry_until_up=retry_until_up,
                skip_unnecessary_provisioning=skip_unnecessary_provisioning)

        if handle is None:
            assert dryrun, ('If not dryrun, handle must be set or '
                            'Stage.PROVISION must be included in stages.')
            logger.info('Dryrun finished.')
            return None, None

        do_workdir = (Stage.SYNC_WORKDIR in stages and not dryrun and
                      task.workdir is not None)
        do_file_mounts = (Stage.SYNC_FILE_MOUNTS in stages and not dryrun and
                          (task.file_mounts is not None or
                           task.storage_mounts is not None))
        if do_workdir or do_file_mounts:
            logger.info(ux_utils.starting_message('Syncing files.'))

        if do_workdir:
            backend.sync_workdir(handle, task.workdir)

        if do_file_mounts:
            backend.sync_file_mounts(handle, task.file_mounts,
                                     task.storage_mounts)

        if no_setup:
            logger.info('Setup commands skipped.')
        elif Stage.SETUP in stages and not dryrun:
            if skip_unnecessary_provisioning and provisioning_skipped:
                logger.debug('Unnecessary provisioning was skipped, so '
                             'skipping setup as well.')
            else:
                backend.setup(handle, task, detach_setup=detach_setup)

        if Stage.PRE_EXEC in stages and not dryrun:
            if idle_minutes_to_autostop is not None:
                assert isinstance(backend, backends.CloudVmRayBackend)
                assert isinstance(handle, backends.CloudVmRayResourceHandle)
                backend.set_autostop(handle,
                                     idle_minutes_to_autostop,
                                     down=down)

        if Stage.EXEC in stages:
            try:
                global_user_state.update_last_use(handle.get_cluster_name())
                job_id = backend.execute(handle,
                                         task,
                                         detach_run,
                                         dryrun=dryrun)
            finally:
                # Enables post_execute() to be run after KeyboardInterrupt.
                backend.post_execute(handle, down)

        if Stage.DOWN in stages and not dryrun:
            if down and idle_minutes_to_autostop is None:
                backend.teardown_ephemeral_storage(task)
                backend.teardown(handle, terminate=True)
    finally:
        print()
        print('\x1b[?25h', end='')  # Show cursor.
    return job_id, handle


@timeline.event
@usage_lib.entrypoint
def launch(
    task: Union['sky.Task', 'sky.Dag'],
    cluster_name: Optional[str] = None,
    retry_until_up: bool = False,
    idle_minutes_to_autostop: Optional[int] = None,
    dryrun: bool = False,
    down: bool = False,
    stream_logs: bool = True,
    backend: Optional[backends.Backend] = None,
    optimize_target: common.OptimizeTarget = common.OptimizeTarget.COST,
    no_setup: bool = False,
    clone_disk_from: Optional[str] = None,
    fast: bool = False,
    # Internal only:
    # pylint: disable=invalid-name
    _quiet_optimizer: bool = False,
    _is_launched_by_jobs_controller: bool = False,
    _is_launched_by_sky_serve_controller: bool = False,
    _disable_controller_check: bool = False,
) -> Tuple[Optional[int], Optional[backends.ResourceHandle]]:
    # NOTE(dev): Keep the docstring consistent between the Python API and CLI.
    """Launches a cluster or task.

    The task's setup and run commands are executed under the task's workdir
    (when specified, it is synced to remote cluster).  The task undergoes job
    queue scheduling on the cluster.

    Currently, the first argument must be a sky.Task, or (EXPERIMENTAL advanced
    usage) a sky.Dag. In the latter case, currently it must contain a single
    task; support for pipelines/general DAGs are in experimental branches.

    Example:
        .. code-block:: python

            import sky
            task = sky.Task(run='echo hello SkyPilot')
            task.set_resources(
                sky.Resources(cloud=sky.AWS(), accelerators='V100:4'))
            sky.launch(task, cluster_name='my-cluster')


    Args:
        task: sky.Task, or sky.Dag (experimental; 1-task only) to launch.
        cluster_name: name of the cluster to create/reuse.  If None,
            auto-generate a name.
        retry_until_up: whether to retry launching the cluster until it is
            up.
        idle_minutes_to_autostop: automatically stop the cluster after this
            many minute of idleness, i.e., no running or pending jobs in the
            cluster's job queue. Idleness gets reset whenever setting-up/
            running/pending jobs are found in the job queue. Setting this
            flag is equivalent to running
            ``sky.launch(...)`` and then
            ``sky.autostop(idle_minutes=<minutes>)``. If not set, the cluster
            will not be autostopped.
        down: Tear down the cluster after all jobs finish (successfully or
            abnormally). If --idle-minutes-to-autostop is also set, the
            cluster will be torn down after the specified idle time.
            Note that if errors occur during provisioning/data syncing/setting
            up, the cluster will not be torn down for debugging purposes.
        dryrun: if True, do not actually launch the cluster.
        stream_logs: if True, show the logs in the terminal.
        backend: backend to use.  If None, use the default backend
            (CloudVMRayBackend).
        optimize_target: target to optimize for. Choices: OptimizeTarget.COST,
            OptimizeTarget.TIME.
        no_setup: if True, do not re-run setup commands.
        clone_disk_from: [Experimental] if set, clone the disk from the
            specified cluster. This is useful to migrate the cluster to a
            different availability zone or region.
        fast: [Experimental] If the cluster is already up and available,
            skip provisioning and setup steps.

    Raises:
        exceptions.ClusterOwnerIdentityMismatchError: if the cluster is
            owned by another user.
        exceptions.InvalidClusterNameError: if the cluster name is invalid.
        exceptions.ResourcesMismatchError: if the requested resources
            do not match the existing cluster.
        exceptions.NotSupportedError: if required features are not supported
            by the backend/cloud/cluster.
        exceptions.ResourcesUnavailableError: if the requested resources
            cannot be satisfied. The failover_history of the exception
            will be set as:
                1. Empty: iff the first-ever sky.optimize() fails to
                find a feasible resource; no pre-check or actual launch is
                attempted.
                2. Non-empty: iff at least 1 exception from either
                our pre-checks (e.g., cluster name invalid) or a region/zone
                throwing resource unavailability.
        exceptions.CommandError: any ssh command error.
        exceptions.NoCloudAccessError: if all clouds are disabled.
    Other exceptions may be raised depending on the backend.

    Returns:
      job_id: Optional[int]; the job ID of the submitted job. None if the
        backend is not CloudVmRayBackend, or no job is submitted to
        the cluster.
      handle: Optional[backends.ResourceHandle]; the handle to the cluster. None
        if dryrun.
    """

    entrypoint = task
    entrypoint.validate()
    if not _disable_controller_check:
        controller_utils.check_cluster_name_not_controller(
            cluster_name, operation_str='sky.launch')

    handle = None
    stages = None
    skip_unnecessary_provisioning = False
    # Check if cluster exists and we are doing fast provisioning
    if fast and cluster_name is not None:
        cluster_status, maybe_handle = (
            backend_utils.refresh_cluster_status_handle(cluster_name))
        if cluster_status == status_lib.ClusterStatus.INIT:
            # If the cluster is INIT, it may be provisioning. We want to prevent
            # concurrent calls from queueing up many sequential reprovision
            # attempts. Since provisioning will hold the cluster status lock, we
            # wait to hold that lock by force refreshing the status. This will
            # block until the cluster finishes provisioning, then correctly see
            # that it is UP.
            # TODO(cooperc): If multiple processes launched in parallel see that
            # the cluster is STOPPED or does not exist, they will still all try
            # to provision it, since we do not hold the lock continuously from
            # the status check until the provision call. Fixing this requires a
            # bigger refactor.
            cluster_status, maybe_handle = (
                backend_utils.refresh_cluster_status_handle(
                    cluster_name,
                    force_refresh_statuses=[
                        # If the cluster is INIT, we want to try to grab the
                        # status lock, which should block until provisioning is
                        # finished.
                        status_lib.ClusterStatus.INIT,
                    ],
                    # Wait indefinitely to obtain the lock, so that we don't
                    # have multiple processes launching the same cluster at
                    # once.
                    cluster_status_lock_timeout=-1,
                ))
        if cluster_status == status_lib.ClusterStatus.UP:
            handle = maybe_handle
            stages = [
                # Provisioning will be short-circuited if the existing
                # cluster config hash matches the calculated one.
                Stage.PROVISION,
                Stage.SYNC_WORKDIR,
                Stage.SYNC_FILE_MOUNTS,
                # Setup will be skipped if provisioning was skipped.
                Stage.SETUP,
                Stage.PRE_EXEC,
                Stage.EXEC,
                Stage.DOWN,
            ]
            skip_unnecessary_provisioning = True

    # Attach to setup if the cluster is a controller, so that user can
    # see the setup logs when inspecting the launch process to know
    # excatly what the job is waiting for.
    detach_setup = controller_utils.Controllers.from_name(cluster_name) is None

    return _execute(
        entrypoint=entrypoint,
        dryrun=dryrun,
        down=down,
        stream_logs=stream_logs,
        handle=handle,
        backend=backend,
        retry_until_up=retry_until_up,
        optimize_target=optimize_target,
        stages=stages,
        cluster_name=cluster_name,
        detach_setup=detach_setup,
        detach_run=True,
        idle_minutes_to_autostop=idle_minutes_to_autostop,
        no_setup=no_setup,
        clone_disk_from=clone_disk_from,
        skip_unnecessary_provisioning=skip_unnecessary_provisioning,
        _quiet_optimizer=_quiet_optimizer,
        _is_launched_by_jobs_controller=_is_launched_by_jobs_controller,
        _is_launched_by_sky_serve_controller=
        _is_launched_by_sky_serve_controller,
    )


@usage_lib.entrypoint
def exec(  # pylint: disable=redefined-builtin
    task: Union['sky.Task', 'sky.Dag'],
    cluster_name: str,
    dryrun: bool = False,
    down: bool = False,
    stream_logs: bool = True,
    backend: Optional[backends.Backend] = None,
) -> Tuple[Optional[int], Optional[backends.ResourceHandle]]:
    # NOTE(dev): Keep the docstring consistent between the Python API and CLI.
    """Executes a task on an existing cluster.

    This function performs two actions:

    (1) workdir syncing, if the task has a workdir specified;
    (2) executing the task's ``run`` commands.

    All other steps (provisioning, setup commands, file mounts syncing) are
    skipped.  If any of those specifications changed in the task, this function
    will not reflect those changes.  To ensure a cluster's setup is up to date,
    use ``sky.launch()`` instead.

    Execution and scheduling behavior:

    - The task will undergo job queue scheduling, respecting any specified
      resource requirement. It can be executed on any node of the cluster with
      enough resources.
    - The task is run under the workdir (if specified).
    - The task is run non-interactively (without a pseudo-terminal or
      pty), so interactive commands such as ``htop`` do not work.
      Use ``ssh my_cluster`` instead.

    Args:
        task: sky.Task, or sky.Dag (experimental; 1-task only) containing the
          task to execute.
        cluster_name: name of an existing cluster to execute the task.
        down: Tear down the cluster after all jobs finish (successfully or
            abnormally). If --idle-minutes-to-autostop is also set, the
            cluster will be torn down after the specified idle time.
            Note that if errors occur during provisioning/data syncing/setting
            up, the cluster will not be torn down for debugging purposes.
        dryrun: if True, do not actually execute the task.
        stream_logs: if True, show the logs in the terminal.
        backend: backend to use.  If None, use the default backend
            (CloudVMRayBackend).

    Raises:
        ValueError: if the specified cluster is not in UP status.
        sky.exceptions.ClusterDoesNotExist: if the specified cluster does not
            exist.
        sky.exceptions.NotSupportedError: if the specified cluster is a
            controller that does not support this operation.

    Returns:
      job_id: Optional[int]; the job ID of the submitted job. None if the
        backend is not CloudVmRayBackend, or no job is submitted to
        the cluster.
      handle: Optional[backends.ResourceHandle]; the handle to the cluster. None
        if dryrun.
    """
    entrypoint = task
    entrypoint.validate(skip_file_mounts=True)
    controller_utils.check_cluster_name_not_controller(cluster_name,
                                                       operation_str='sky.exec')

    handle = backend_utils.check_cluster_available(
        cluster_name,
        operation='executing tasks',
        check_cloud_vm_ray_backend=False,
        dryrun=dryrun)
    return _execute(
        entrypoint=entrypoint,
        dryrun=dryrun,
        down=down,
        stream_logs=stream_logs,
        handle=handle,
        backend=backend,
        stages=[
            Stage.SYNC_WORKDIR,
            Stage.EXEC,
        ],
        cluster_name=cluster_name,
        detach_run=True,
    )



#!/usr/bin/env python3
# /// script
# dependencies = [
#   "websockets>=14.0",
# ]
# ///
"""Starting a websocket with SkyPilot API server to proxy SSH to a k8s pod.

This script is useful for users who do not have local Kubernetes credentials.
"""
import asyncio
from http.cookiejar import MozillaCookieJar
import os
import sys
from typing import Dict
from urllib.request import Request

import websockets
from websockets.asyncio.client import connect


def _get_cookie_header(url: str) -> Dict[str, str]:
    """Extract Cookie header value from a cookie jar for a specific URL"""
    cookie_path = os.environ.get('SKYPILOT_API_COOKIE_FILE')
    if cookie_path is None:
        return {}

    request = Request(url)
    cookie_jar = MozillaCookieJar(os.path.expanduser(cookie_path))
    cookie_jar.load(ignore_discard=True, ignore_expires=True)
    cookie_jar.add_cookie_header(request)
    cookie_header = request.get_header('Cookie')
    # if cookie file is empty, return empty dict
    if cookie_header is None:
        return {}
    return {'Cookie': cookie_header}


async def main(url: str) -> None:
    cookie_header = _get_cookie_header(url)
    async with connect(url,
                       ping_interval=None,
                       additional_headers=cookie_header) as websocket:
        if os.isatty(sys.stdin.fileno()):
            # pylint: disable=import-outside-toplevel
            import termios
            import tty
            old_settings = termios.tcgetattr(sys.stdin.fileno())
            tty.setraw(sys.stdin.fileno())
        else:
            old_settings = None

        try:
            await asyncio.gather(stdin_to_websocket(websocket),
                                 websocket_to_stdout(websocket))
        finally:
            if old_settings:
                termios.tcsetattr(sys.stdin.fileno(), termios.TCSADRAIN,
                                  old_settings)


async def stdin_to_websocket(websocket):
    try:
        while True:
            data = await asyncio.get_event_loop().run_in_executor(
                None, sys.stdin.buffer.read, 1)
            if not data:
                break
            await websocket.send(data)
    except Exception as e:  # pylint: disable=broad-except
        print(f'Error in stdin_to_websocket: {e}', file=sys.stderr)
    finally:
        await websocket.close()


async def websocket_to_stdout(websocket):
    try:
        while True:
            message = await websocket.recv()
            sys.stdout.buffer.write(message)
            await asyncio.get_event_loop().run_in_executor(
                None, sys.stdout.buffer.flush)
    except websockets.exceptions.ConnectionClosed:
        print('WebSocket connection closed', file=sys.stderr)
    except Exception as e:  # pylint: disable=broad-except
        print(f'Error in websocket_to_stdout: {e}', file=sys.stderr)


if __name__ == '__main__':
    server_url = sys.argv[1].strip('/')
    if '://' not in server_url:
        # Keep backward compatibility for legacy server URLs without protocol
        # TODO(aylei): Remove this after 0.10.0
        server_url = f'http://{server_url}'

    server_proto, server_fqdn = server_url.split('://')
    websocket_proto = 'ws'
    if server_proto == 'https':
        websocket_proto = 'wss'
    server_url = f'{websocket_proto}://{server_fqdn}'
    websocket_url = (f'{server_url}/kubernetes-pod-ssh-proxy'
                     f'?cluster_name={sys.argv[2]}')
    asyncio.run(main(websocket_url))



"""Credential checks: check cloud credentials and enable clouds."""
import os
import traceback
from types import ModuleType
from typing import (Any, Callable, Dict, Iterable, List, Optional, Set, Tuple,
                    Union)

import click
import colorama

from sky import clouds as sky_clouds
from sky import exceptions
from sky import global_user_state
from sky import skypilot_config
from sky.adaptors import cloudflare
from sky.clouds import cloud as sky_cloud
from sky.utils import registry
from sky.utils import rich_utils
from sky.utils import ux_utils

CHECK_MARK_EMOJI = '\U00002714'  # Heavy check mark unicode
PARTY_POPPER_EMOJI = '\U0001F389'  # Party popper unicode


def check_capabilities(
    quiet: bool = False,
    verbose: bool = False,
    clouds: Optional[Iterable[str]] = None,
    capabilities: Optional[List[sky_cloud.CloudCapability]] = None,
) -> Dict[str, List[sky_cloud.CloudCapability]]:
    echo = (lambda *_args, **_kwargs: None
           ) if quiet else lambda *args, **kwargs: click.echo(
               *args, **kwargs, color=True)
    echo('Checking credentials to enable clouds for SkyPilot.')
    if capabilities is None:
        capabilities = sky_cloud.ALL_CAPABILITIES
    assert capabilities is not None
    enabled_clouds: Dict[str, List[sky_cloud.CloudCapability]] = {}
    disabled_clouds: Dict[str, List[sky_cloud.CloudCapability]] = {}

    def check_one_cloud(
            cloud_tuple: Tuple[str, Union[sky_clouds.Cloud,
                                          ModuleType]]) -> None:
        cloud_repr, cloud = cloud_tuple
        assert capabilities is not None
        # cloud_capabilities is a list of (capability, ok, reason)
        # where ok is True if the cloud credentials are valid for the capability
        cloud_capabilities: List[Tuple[sky_cloud.CloudCapability, bool,
                                       Optional[str]]] = []
        for capability in capabilities:
            with rich_utils.safe_status(f'Checking {cloud_repr}...'):
                try:
                    ok, reason = cloud.check_credentials(capability)
                except exceptions.NotSupportedError:
                    continue
                except Exception:  # pylint: disable=broad-except
                    # Catch all exceptions to prevent a single cloud
                    # from blocking the check for other clouds.
                    ok, reason = False, traceback.format_exc()
            cloud_capabilities.append(
                (capability, ok, reason.strip() if reason else None))
            if ok:
                enabled_clouds.setdefault(cloud_repr, []).append(capability)
            else:
                disabled_clouds.setdefault(cloud_repr, []).append(capability)
        _print_checked_cloud(echo, verbose, cloud_tuple, cloud_capabilities)

    def get_cloud_tuple(
            cloud_name: str) -> Tuple[str, Union[sky_clouds.Cloud, ModuleType]]:
        # Validates cloud_name and returns a tuple of the cloud's name and
        # the cloud object. Includes special handling for Cloudflare.
        if cloud_name.lower().startswith('cloudflare'):
            return cloudflare.NAME, cloudflare
        else:
            cloud_obj = registry.CLOUD_REGISTRY.from_str(cloud_name)
            assert cloud_obj is not None, f'Cloud {cloud_name!r} not found'
            return repr(cloud_obj), cloud_obj

    def get_all_clouds():
        return tuple([repr(c) for c in registry.CLOUD_REGISTRY.values()] +
                     [cloudflare.NAME])

    if clouds is not None:
        cloud_list = clouds
    else:
        cloud_list = get_all_clouds()
    clouds_to_check = [get_cloud_tuple(c) for c in cloud_list]

    # Use allowed_clouds from config if it exists, otherwise check all clouds.
    # Also validate names with get_cloud_tuple.
    config_allowed_cloud_names = [
        get_cloud_tuple(c)[0] for c in skypilot_config.get_nested((
            'allowed_clouds',), get_all_clouds())
    ]
    # Use disallowed_cloud_names for logging the clouds that will be disabled
    # because they are not included in allowed_clouds in config.yaml.
    disallowed_cloud_names = [
        c for c in get_all_clouds() if c not in config_allowed_cloud_names
    ]
    # Check only the clouds which are allowed in the config.
    clouds_to_check = [
        c for c in clouds_to_check if c[0] in config_allowed_cloud_names
    ]

    for cloud_tuple in sorted(clouds_to_check):
        check_one_cloud(cloud_tuple)

    # Determine the set of enabled clouds: (previously enabled clouds + newly
    # enabled clouds - newly disabled clouds) intersected with
    # config_allowed_clouds, if specified in config.yaml.
    # This means that if a cloud is already enabled and is not included in
    # allowed_clouds in config.yaml, it will be disabled.
    all_enabled_clouds: Set[str] = set()
    for capability in capabilities:
        # Cloudflare is not a real cloud in registry.CLOUD_REGISTRY, and should
        # not be inserted into the DB (otherwise `sky launch` and other code
        # would error out when it's trying to look it up in the registry).
        enabled_clouds_set = {
            cloud for cloud, capabilities in enabled_clouds.items()
            if capability in capabilities and not cloud.startswith('Cloudflare')
        }
        disabled_clouds_set = {
            cloud for cloud, capabilities in disabled_clouds.items()
            if capability in capabilities and not cloud.startswith('Cloudflare')
        }
        config_allowed_clouds_set = {
            cloud for cloud in config_allowed_cloud_names
            if not cloud.startswith('Cloudflare')
        }
        previously_enabled_clouds_set = {
            repr(cloud)
            for cloud in global_user_state.get_cached_enabled_clouds(capability)
        }
        enabled_clouds_for_capability = (config_allowed_clouds_set & (
            (previously_enabled_clouds_set | enabled_clouds_set) -
            disabled_clouds_set))
        global_user_state.set_enabled_clouds(
            list(enabled_clouds_for_capability), capability)
        all_enabled_clouds = all_enabled_clouds.union(
            enabled_clouds_for_capability)
    disallowed_clouds_hint = None
    if disallowed_cloud_names:
        disallowed_clouds_hint = (
            '\nNote: The following clouds were disabled because they were not '
            'included in allowed_clouds in ~/.sky/config.yaml: '
            f'{", ".join([c for c in disallowed_cloud_names])}')
    if not all_enabled_clouds:
        echo(
            click.style(
                'No cloud is enabled. SkyPilot will not be able to run any '
                'task. Run `sky check` for more info.',
                fg='red',
                bold=True))
        if disallowed_clouds_hint:
            echo(click.style(disallowed_clouds_hint, dim=True))
        raise SystemExit()
    else:
        clouds_arg = (f' {" ".join(disabled_clouds).lower()}'
                      if clouds is not None else '')
        echo(
            click.style(
                '\nTo enable a cloud, follow the hints above and rerun: ',
                dim=True) + click.style(f'sky check{clouds_arg}', bold=True) +
            '\n' + click.style(
                'If any problems remain, refer to detailed docs at: '
                'https://docs.skypilot.co/en/latest/getting-started/installation.html',  # pylint: disable=line-too-long
                dim=True))

        if disallowed_clouds_hint:
            echo(click.style(disallowed_clouds_hint, dim=True))

        # Pretty print for UX.
        if not quiet:
            enabled_clouds_str = '\n  ' + '\n  '.join([
                _format_enabled_cloud(cloud, capabilities)
                for cloud, capabilities in enabled_clouds.items()
            ])
            echo(f'\n{colorama.Fore.GREEN}{PARTY_POPPER_EMOJI} '
                 f'Enabled clouds {PARTY_POPPER_EMOJI}'
                 f'{colorama.Style.RESET_ALL}{enabled_clouds_str}')
    return enabled_clouds


def check_capability(
    capability: sky_cloud.CloudCapability,
    quiet: bool = False,
    verbose: bool = False,
    clouds: Optional[Iterable[str]] = None,
) -> List[str]:
    clouds_with_capability = []
    enabled_clouds = check_capabilities(quiet, verbose, clouds, [capability])
    for cloud, capabilities in enabled_clouds.items():
        if capability in capabilities:
            clouds_with_capability.append(cloud)
    return clouds_with_capability


def check(
    quiet: bool = False,
    verbose: bool = False,
    clouds: Optional[Iterable[str]] = None,
) -> List[str]:
    return list(
        check_capabilities(quiet, verbose, clouds,
                           sky_cloud.ALL_CAPABILITIES).keys())


def get_cached_enabled_clouds_or_refresh(
        capability: sky_cloud.CloudCapability,
        raise_if_no_cloud_access: bool = False) -> List[sky_clouds.Cloud]:
    """Returns cached enabled clouds and if no cloud is enabled, refresh.

    This function will perform a refresh if no public cloud is enabled.

    Args:
        raise_if_no_cloud_access: if True, raise an exception if no public
            cloud is enabled.

    Raises:
        exceptions.NoCloudAccessError: if no public cloud is enabled and
            raise_if_no_cloud_access is set to True.
    """
    cached_enabled_clouds = global_user_state.get_cached_enabled_clouds(
        capability)
    if not cached_enabled_clouds:
        try:
            check_capability(sky_cloud.CloudCapability.COMPUTE, quiet=True)
        except SystemExit:
            # If no cloud is enabled, check() will raise SystemExit.
            # Here we catch it and raise the exception later only if
            # raise_if_no_cloud_access is set to True.
            pass
        cached_enabled_clouds = global_user_state.get_cached_enabled_clouds(
            capability)
    if raise_if_no_cloud_access and not cached_enabled_clouds:
        with ux_utils.print_exception_no_traceback():
            raise exceptions.NoCloudAccessError(
                'Cloud access is not set up. Run: '
                f'{colorama.Style.BRIGHT}sky check{colorama.Style.RESET_ALL}')
    return cached_enabled_clouds


def get_cloud_credential_file_mounts(
        excluded_clouds: Optional[Iterable[sky_clouds.Cloud]]
) -> Dict[str, str]:
    """Returns the files necessary to access all clouds.

    Returns a dictionary that will be added to a task's file mounts
    and a list of patterns that will be excluded (used as rsync_exclude).
    """
    # Uploading credentials for all clouds instead of only sky check
    # enabled clouds because users may have partial credentials for some
    # clouds to access their specific resources (e.g. cloud storage) but
    # not have the complete credentials to pass sky check.
    clouds = registry.CLOUD_REGISTRY.values()
    file_mounts = {}
    for cloud in clouds:
        if (excluded_clouds is not None and
                sky_clouds.cloud_in_iterable(cloud, excluded_clouds)):
            continue
        cloud_file_mounts = cloud.get_credential_file_mounts()
        for remote_path, local_path in cloud_file_mounts.items():
            if os.path.exists(os.path.expanduser(local_path)):
                file_mounts[remote_path] = local_path
    # Currently, get_cached_enabled_clouds_or_refresh() does not support r2 as
    # only clouds with computing instances are marked as enabled by skypilot.
    # This will be removed when cloudflare/r2 is added as a 'cloud'.
    r2_is_enabled, _ = cloudflare.check_storage_credentials()
    if r2_is_enabled:
        r2_credential_mounts = cloudflare.get_credential_file_mounts()
        file_mounts.update(r2_credential_mounts)
    return file_mounts


def _print_checked_cloud(
    echo: Callable,
    verbose: bool,
    cloud_tuple: Tuple[str, Union[sky_clouds.Cloud, ModuleType]],
    cloud_capabilities: List[Tuple[sky_cloud.CloudCapability, bool,
                                   Optional[str]]],
) -> None:
    """Prints whether a cloud is enabled, and the capabilities that are enabled.
    If any hints (for enabled capabilities) or
    reasons (for disabled capabilities) are provided, they will be printed.

    Args:
        echo: The function to use to print the message.
        verbose: Whether to print the verbose output.
        cloud_tuple: The cloud to print the capabilities for.
        cloud_capabilities: The capabilities for the cloud.
    """

    def _yellow_color(str_to_format: str) -> str:
        return (f'{colorama.Fore.LIGHTYELLOW_EX}'
                f'{str_to_format}'
                f'{colorama.Style.RESET_ALL}')

    cloud_repr, cloud = cloud_tuple
    # Print the capabilities for the cloud.
    # consider cloud enabled if any capability is enabled.
    enabled_capabilities: List[sky_cloud.CloudCapability] = []
    hints_to_capabilities: Dict[str, List[sky_cloud.CloudCapability]] = {}
    reasons_to_capabilities: Dict[str, List[sky_cloud.CloudCapability]] = {}
    for capability, ok, reason in cloud_capabilities:
        if ok:
            enabled_capabilities.append(capability)
            if reason is not None:
                hints_to_capabilities.setdefault(reason, []).append(capability)
        elif reason is not None:
            reasons_to_capabilities.setdefault(reason, []).append(capability)
    status_msg: str = 'disabled'
    styles: Dict[str, Any] = {'dim': True}
    capability_string: str = ''
    activated_account: Optional[str] = None
    if enabled_capabilities:
        status_msg = 'enabled'
        styles = {'fg': 'green', 'bold': False}
        capability_string = f'[{", ".join(enabled_capabilities)}]'
        if verbose and cloud is not cloudflare:
            activated_account = cloud.get_active_user_identity_str()

    echo(
        click.style(f'  {cloud_repr}: {status_msg} {capability_string}',
                    **styles))
    if activated_account is not None:
        echo(f'    Activated account: {activated_account}')
    for reason, caps in hints_to_capabilities.items():
        echo(f'    Hint [{", ".join(caps)}]: {_yellow_color(reason)}')
    for reason, caps in reasons_to_capabilities.items():
        echo(f'    Reason [{", ".join(caps)}]: {reason}')


def _format_enabled_cloud(cloud_name: str,
                          capabilities: List[sky_cloud.CloudCapability]) -> str:
    """Format the summary of enabled cloud and its enabled capabilities.

    Args:
        cloud_name: The name of the cloud.
        capabilities: The capabilities of the cloud.

    Returns:
        A string of the formatted cloud and capabilities.
    """
    cloud_and_capabilities = f'{cloud_name} [{", ".join(capabilities)}]'

    def _green_color(str_to_format: str) -> str:
        return (
            f'{colorama.Fore.GREEN}{str_to_format}{colorama.Style.RESET_ALL}')

    if cloud_name == repr(sky_clouds.Kubernetes()):
        # Get enabled contexts for Kubernetes
        existing_contexts = sky_clouds.Kubernetes.existing_allowed_contexts()
        if not existing_contexts:
            return _green_color(cloud_and_capabilities)

        # Check if allowed_contexts is explicitly set in config
        allowed_contexts = skypilot_config.get_nested(
            ('kubernetes', 'allowed_contexts'), None)

        # Format the context info with consistent styling
        if allowed_contexts is not None:
            contexts_formatted = []
            for i, context in enumerate(existing_contexts):
                symbol = (ux_utils.INDENT_LAST_SYMBOL
                          if i == len(existing_contexts) -
                          1 else ux_utils.INDENT_SYMBOL)
                contexts_formatted.append(f'\n    {symbol}{context}')
            context_info = f'  Allowed contexts:{"".join(contexts_formatted)}'
        else:
            context_info = f'  Active context: {existing_contexts[0]}'

        return (f'{_green_color(cloud_and_capabilities)}\n'
                f'  {colorama.Style.DIM}{context_info}'
                f'{colorama.Style.RESET_ALL}')
    return _green_color(cloud_and_capabilities)



"""CLS API Helper
"""

import os
import ssl
import time
import urllib.request as urllib2

from sky import sky_logging
from sky.adaptors import vsphere as vsphere_adaptor
from sky.provision.vsphere.common.cls_api_client import ClsApiClient
from sky.provision.vsphere.common.id_generator import generate_random_uuid
from sky.provision.vsphere.common.vim_utils import get_obj_by_mo_id

logger = sky_logging.init_logger(__name__)


class ClsApiHelper(object):
    """Helper class to perform commonly used operations
    using Content Library API.
    """

    ISO_FILE_RELATIVE_DIR = '../resources/isoImages/'
    PLAIN_OVF_RELATIVE_DIR = '../resources/plainVmTemplate'
    SIMPLE_OVF_RELATIVE_DIR = '../resources/simpleVmTemplate'

    def __init__(self, cls_api_client: ClsApiClient, skip_verification: bool):
        self.client = cls_api_client
        self.skip_verification = skip_verification

    def get_ovf_files_map(self, ovf_location):
        """Get OVF template file paths to be used during uploads
        Note: This method returns OVF template paths for the template included
              in the SDK resources directory
        """

        ovf_files_map = {}
        ovf_dir = os.path.abspath(
            os.path.join(os.path.dirname(os.path.realpath(__file__)),
                         ovf_location))
        for file_name in os.listdir(ovf_dir):
            if file_name.endswith('.ovf') or file_name.endswith('.vmdk'):
                ovf_files_map[file_name] = os.path.join(ovf_dir, file_name)
        return ovf_files_map

    def create_local_library(self, storage_backings, lib_name):
        """:param storage_backings: Storage for the library
        :param lib_name: Name of the library
        :return: id of the created library
        """
        create_spec = vsphere_adaptor.get_content_client().LibraryModel()
        create_spec.name = lib_name
        create_spec.description = 'Local library backed by VC datastore'
        create_spec.type = vsphere_adaptor.get_content_client(
        ).LibraryModel.LibraryType.LOCAL
        create_spec.storage_backings = storage_backings

        # Create a local content library backed the VC datastore
        library_id = self.client.local_library_service.create(
            create_spec=create_spec, client_token=generate_random_uuid())
        logger.info('Local library created, ID: {0}'.format(library_id))

        return library_id

    def create_iso_library_item(self, library_id, iso_item_name, iso_filename):
        """:param library_id: item will be created on this library
        :param iso_item_name: name of the iso item to be created
        :param iso_filename: name of the iso file to be uploaded
        :return: id of the item created
        """
        # Create a new library item in the content library for uploading the
        # files
        library_item_id = self.create_library_item(
            library_id=library_id,
            item_name=iso_item_name,
            item_description='Sample iso file',
            item_type='iso',
        )
        assert library_item_id is not None
        logger.info('Library item created id: {0}'.format(library_item_id))

        # Upload an iso file to above library item, use the filename as the
        # item_filename
        iso_files_map = self.get_iso_file_map(item_filename=iso_filename,
                                              disk_filename=iso_filename)
        self.upload_files(library_item_id=library_item_id,
                          files_map=iso_files_map)
        logger.info(
            'Uploaded iso file to library item {0}'.format(library_item_id))
        return library_item_id

    def get_iso_file_map(self, item_filename, disk_filename):
        iso_files_map = {}
        iso_file_path = os.path.abspath(
            os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                self.ISO_FILE_RELATIVE_DIR + disk_filename,
            ))
        iso_files_map[item_filename] = iso_file_path
        return iso_files_map

    def get_ova_file_map(self, relative_path, local_filename):
        """Get OVA file paths to be used during uploads.

        :param relative_path: directory path under contentlibrary/resources
         for ova file
        :param local_filename: name of the file on local disk under resources
         directory
        :return: mapping of item's filename to full path for the file on local
         disk
        """
        ova_file_map = {}
        item_filename = os.path.basename(local_filename)
        ova_file_path = os.path.abspath(
            os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                relative_path,
                local_filename,
            ))
        ova_file_map[item_filename] = ova_file_path
        return ova_file_map

    def get_libraryitem_spec(
            self,
            client_token,  # pylint: disable=unused-argument
            name,
            description,
            library_id,  # pylint: disable=unused-argument
            library_item_type):
        """        Create library item spec

        """
        lib_item_spec = vsphere_adaptor.get_library_client().ItemModel()
        lib_item_spec.name = name
        lib_item_spec.description = description
        lib_item_spec.library_id = library_id
        lib_item_spec.type = library_item_type
        return lib_item_spec

    def create_library_item(self, library_id, item_name, item_description,
                            item_type):
        """Create a library item in the specified library
        """
        lib_item_spec = self.get_libraryitem_spec(
            client_token=generate_random_uuid(),
            name=item_name,
            description=item_description,
            library_id=library_id,
            library_item_type=item_type,
        )
        # Create a library item
        return self.client.library_item_service.create(
            create_spec=lib_item_spec, client_token=generate_random_uuid())

    def upload_files(self, library_item_id, files_map):
        """Upload a VM template to the published CL
        """
        # Create a new upload session for uploading the files
        session_id = self.client.upload_service.create(
            create_spec=vsphere_adaptor.get_item_client().UpdateSessionModel(
                library_item_id=library_item_id),
            client_token=generate_random_uuid(),
        )
        self.upload_files_in_session(files_map, session_id)
        self.client.upload_service.complete(session_id)
        self.client.upload_service.delete(session_id)

    def upload_files_in_session(self, files_map, session_id):
        for f_name, f_path in files_map.items():
            file_spec = self.client.upload_file_service.AddSpec(
                name=f_name,
                source_type=vsphere_adaptor.get_updatesession_client().File.
                SourceType.PUSH,
                size=os.path.getsize(f_path),
            )
            file_info = self.client.upload_file_service.add(
                session_id, file_spec)
            # Upload the file content to the file upload URL
            with open(f_path, 'rb') as local_file:
                request = urllib2.Request(file_info.upload_endpoint.uri,
                                          local_file)
                request.add_header('Cache-Control', 'no-cache')
                request.add_header('Content-Length',
                                   '{0}'.format(os.path.getsize(f_path)))
                request.add_header('Content-Type', 'text/ovf')
                if self.skip_verification and hasattr(
                        ssl, '_create_unverified_context'):
                    # Python 2.7.9 has stronger SSL certificate validation,
                    # so we need to pass in a context when dealing with
                    # self-signed certificates.
                    context = ssl._create_unverified_context()  # pylint: disable=protected-access
                    urllib2.urlopen(request, context=context)
                else:
                    # Don't pass context parameter since versions of Python
                    # before 2.7.9 don't support it.
                    urllib2.urlopen(request)

    def download_files(self, library_item_id, directory):
        """Download files from a library item

        Args:
            library_item_id: id for the library item to download files from
            directory: location on the client machine to download the files into

        """
        downloaded_files_map = {}
        # create a new download session for downloading the session files
        session_id = self.client.download_service.create(
            create_spec=vsphere_adaptor.get_item_client().DownloadSessionModel(
                library_item_id=library_item_id),
            client_token=generate_random_uuid(),
        )
        file_infos = self.client.download_file_service.list(session_id)
        for file_info in file_infos:
            self.client.download_file_service.prepare(session_id,
                                                      file_info.name)
            download_info = self.wait_for_prepare(session_id, file_info.name)
            if self.skip_verification and hasattr(ssl,
                                                  '_create_unverified_context'):
                # Python 2.7.9 has stronger SSL certificate validation,
                # so we need to pass in a context when dealing with self-signed
                # certificates.
                context = ssl._create_unverified_context()  # pylint: disable=protected-access
                response = urllib2.urlopen(  # pylint: disable=consider-using-with
                    url=download_info.download_endpoint.uri,
                    context=context)
            else:
                # Don't pass context parameter since versions of Python
                # before 2.7.9 don't support it.
                response = urllib2.urlopen(download_info.download_endpoint.uri)  # pylint: disable=consider-using-with
            file_path = os.path.join(directory, file_info.name)
            with open(file_path, 'wb') as local_file:
                local_file.write(response.read())
            downloaded_files_map[file_info.name] = file_path
        self.client.download_service.delete(session_id)
        return downloaded_files_map

    def wait_for_prepare(
        self,
        session_id,
        file_name,
        status_list=None,
        timeout=30,
        sleep_interval=1,
    ):
        """        Waits for a file to reach a status in the status list
         (default: prepared)
        This method will either timeout or return the result of
        downloadSessionFile.get(session_id, file_name)

        """
        if status_list is None:
            # Set default status list to prepared
            status_list = [
                vsphere_adaptor.get_updatesession_client().File.Status.PREPARED
            ]
        start_time = time.time()
        while (time.time() - start_time) < timeout:
            file_info = self.client.download_file_service.get(
                session_id, file_name)
            if file_info.status in status_list:
                return file_info
            else:
                time.sleep(sleep_interval)
        raise Exception(
            'timed out after waiting {0} seconds for file {1} to reach'
            ' a terminal state'.format(timeout, file_name))

    def get_item_id_by_name(self, name):
        """        Returns the identifier of the item with the given name.

        Args:
            name (str): The name of item to look for

        Returns:
            str: The item ID or None if the item is not found
        """
        find_spec = vsphere_adaptor.get_library_client().Item.FindSpec(
            name=name)
        item_ids = self.client.library_item_service.find(find_spec)
        item_id = item_ids[0] if item_ids else None
        if item_id:
            logger.info('Library item ID: {0}'.format(item_id))
        else:
            logger.info('Library item with name \'{0}\' not found'.format(name))
        return item_id

    def deploy_ovf_template(self, vm_name, lib_item_id, ovf_summary,
                            deployment_target, storage_profile_id):
        # Build the deployment spec
        deployment_spec = vsphere_adaptor.get_ovf_client(
        ).LibraryItem.ResourcePoolDeploymentSpec(
            name=vm_name,
            annotation=ovf_summary.annotation,
            accept_all_eula=True,
            network_mappings=None,
            storage_mappings=None,
            storage_provisioning=None,
            storage_profile_id=storage_profile_id,
            locale=None,
            flags=None,
            additional_parameters=None,
        )

        # Deploy the ovf template
        result = self.client.ovf_lib_item_service.deploy(
            lib_item_id,
            deployment_target,
            deployment_spec,
            client_token=generate_random_uuid(),
        )

        # The type and ID of the target deployment is available in the
        # deployment result.
        if result.succeeded:
            logger.info(
                'Deployment successful. Result resource: {0}, ID: {1}'.format(
                    result.resource_id.type, result.resource_id.id))
            self.vm_id = result.resource_id.id
            error = result.error
            if error is not None:
                for warning in error.warnings:
                    logger.info('OVF warning: {}'.format(warning.message))

            # Power on the VM and wait  for the power on operation to be
            # completed
            self.vm_obj = get_obj_by_mo_id(
                self.client.service_manager.content,
                [vsphere_adaptor.get_vim().VirtualMachine], self.vm_id)
            assert self.vm_obj is not None
            return self.vm_obj
        else:
            logger.info('Deployment failed.')
            for error in result.error.errors:
                logger.info('OVF error: {}'.format(error.message))
